"""Classes concerning data schemas."""
from __future__ import annotations

import collections
import copy
import logging
from os import PathLike
from typing import (
    Any,
    Dict,
    List,
    Mapping,
    MutableMapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
    cast,
)

from PIL import Image
from marshmallow import fields, post_dump, post_load, pre_dump, pre_load
import numpy as np
import pandas as pd
from pandas._typing import Dtype
from pandas.core.dtypes.common import is_numeric_dtype
import yaml

import bitfount
from bitfount.data import BaseSource
from bitfount.data.datasources.base_source import MultiTableSource
from bitfount.data.exceptions import BitfountSchemaError, DataSourceError
from bitfount.data.types import (
    CategoricalRecord,
    ContinuousRecord,
    ImageRecord,
    SemanticType,
    TextRecord,
    _CamelCaseSchema,
    _FeatureDict,
    _SemanticTypeRecord,
    _SemanticTypeValue,
)
from bitfount.data.utils import _hash_str
from bitfount.types import _Dtypes, _JSONDict
from bitfount.utils import _add_this_to_list, delegates

logger = logging.getLogger(__name__)


class _TableSchemaMarshmallowMixIn:
    class _Schema(_CamelCaseSchema):
        name = fields.Str(required=True)
        description = fields.Str(allow_none=True)
        categorical_features = fields.List(fields.Nested(CategoricalRecord._Schema))
        continuous_features = fields.List(fields.Nested(ContinuousRecord._Schema))
        image_features = fields.List(fields.Nested(ImageRecord._Schema))
        text_features = fields.List(fields.Nested(TextRecord._Schema))

        @pre_dump
        def dump_feature_values(self, data: TableSchema, **_kwargs: Any) -> TableSchema:
            """Modifies features to dump features as a list instead of dictionaries.

            To ensure our YAML is clear, we pre-process our object into lists before
            dumping it. We don't want to modify the actual schema object, as it will
            affect its use, so we create a temporary one just for dumping to YAML.
            """
            temp_schema = copy.deepcopy(data)
            for stype in data.features:
                setattr(
                    temp_schema,
                    f"{stype}_features",
                    list(data.features[cast(_SemanticTypeValue, stype)].values()),
                )

            return temp_schema

        @post_dump
        def combine_features(self, data: _JSONDict, **kwargs: Any) -> _JSONDict:
            """Combines features belonging to different semantic types under one key.

            After combining the features into one list, it also sorts all the features
            by featureName.
            """
            new_data = {}
            new_data["name"] = data.get("name")
            new_data["description"] = data.get("description")
            features: List[_JSONDict] = [
                item for key in data if key.endswith("Features") for item in data[key]
            ]
            # sort features alphabetically
            # Type ignore due to this bug: https://github.com/python/mypy/issues/9656
            new_data["features"] = sorted(features, key=lambda d: d["featureName"])  # type: ignore[no-any-return] # Reason: see comment above # noqa: B950
            return new_data

        @pre_load
        def split_features(self, data: _JSONDict, **kwargs: Any) -> _JSONDict:
            """Splits features back into a dictionary of lists by semantic type."""
            result = collections.defaultdict(list)
            if "features" in data:
                # Workaround to ensure that the data is not pre-processed
                # twice for the bitfount reference model.
                features: List[_JSONDict] = data.pop("features")
                for d in features:
                    result[d.pop("semanticType")].append(d)

                for semantic_type in result:
                    data[f"{semantic_type}Features"] = result[semantic_type]
                return data
            elif any([key for key in data if "Features" in key]):
                # Data has been already preprocessed
                return data
            else:
                raise ValueError("No features found in the schema.")

        @post_load
        def recreate_schema(self, data: _JSONDict, **_kwargs: Any) -> TableSchema:
            """Recreates Schema."""
            new_schema = TableSchema(
                name=data["name"], description=data.get("description")
            )

            for key in data:
                if key.endswith("_features"):
                    stype = key.replace("_features", "")
                    new_schema.features[cast(_SemanticTypeValue, stype)] = {
                        feature.feature_name: feature for feature in data[key]
                    }

            return new_schema


@delegates()
class TableSchema(_TableSchemaMarshmallowMixIn):
    """A schema that defines the features of a dataframe.

    It lists all the (categorical, continuous, image, and
    text) features found in the dataframe.

    Args:
        name: The name of the table.
        description: A description of the table.

    Attributes:
        name: The name of the table.
        description: A description of the table. Optional.
        features: An ordered dictionary of features (column names).
    """

    def __init__(self, name: str, description: Optional[str] = None):
        # ordered dictionaries of features (column names)
        self.name: str = name
        self.description: Optional[str] = description
        self.features: _FeatureDict = _FeatureDict()

    @staticmethod
    def _dtype_based_stype_split(
        data: _Dtypes, ignore_cols: Optional[Sequence[str]] = None
    ) -> Dict[SemanticType, List[str]]:
        """Returns dictionary of Semantic types and corresponding columns in `data`.

        This method determines which data types correspond to which semantic types.
        """
        converted_data = data.copy()
        if ignore_cols:
            missing_cols = [i for i in ignore_cols if i not in data]
            if missing_cols:
                logger.warning(
                    f"Could not find ignored columns: {', '.join(missing_cols)}"
                )
            converted_data = {
                k: v for k, v in converted_data.items() if k not in ignore_cols
            }

        semantic_types: Dict[SemanticType, List[str]] = {
            stype: [] for stype in SemanticType
        }

        for col, typ in converted_data.items():
            if isinstance(typ, pd.StringDtype) or typ == str:
                semantic_types[SemanticType.TEXT].append(col)
            elif isinstance(typ, pd.BooleanDtype) or typ == bool:
                semantic_types[SemanticType.CATEGORICAL].append(col)
            elif is_numeric_dtype(typ):
                # Booleans get interpereted as continuous so we must define them as
                # categorical before this function is called
                semantic_types[SemanticType.CONTINUOUS].append(col)
            else:
                # By default everything else will be interpreted as categorical.
                # This should only happen for columns which remain as `object` because
                # pandas is having trouble deciphering their true type
                semantic_types[SemanticType.CATEGORICAL].append(col)

        return {k: v for k, v in semantic_types.items() if len(v) > 0}

    def _add_categorical_feature(
        self,
        name: str,
        values: Union[np.ndarray, pd.Series],
        dtype: Optional[Union[Dtype, np.dtype]] = None,
        description: Optional[str] = None,
    ) -> None:
        """Adds the given categorical, with list of values to the schema."""
        if (
            "categorical" not in self.features
            or name not in self.features["categorical"]
        ):

            CategoricalRecord.add_record_to_schema(
                self,
                feature_name=name,
                dtype=dtype,
                description=description,
            )
        self.features["categorical"][name].encoder.add_values(values)

    def _combine_existing_stypes_with_forced_stypes(
        self,
        existing_stypes: MutableMapping[SemanticType, List[str]],
        forced_stype: MutableMapping[_SemanticTypeValue, List[str]],
        table_dtypes: Mapping[str, Any],
    ) -> MutableMapping[SemanticType, List[str]]:
        """Combine the exiting semantic types with the forced semantic types."""
        for new_stype, feature_list in forced_stype.items():
            try:
                stype = SemanticType(new_stype)

                if stype not in existing_stypes:
                    existing_stypes[stype] = []
                existing_stypes[stype] = _add_this_to_list(
                    feature_list, existing_stypes[stype]
                )
            except ValueError:
                logger.warning(
                    f"Given semantic type {new_stype} is not currently supported. "
                    f"Defaulting to split based on dtype."
                )
                feature_dtypes = {
                    k: v for k, v in table_dtypes.items() if k in feature_list
                }
                dtype_features = self._dtype_based_stype_split(feature_dtypes, [])
                stype = list(dtype_features)[0]
                if stype not in existing_stypes:
                    existing_stypes[stype] = []
                existing_stypes[stype] = _add_this_to_list(
                    feature_list, existing_stypes[stype]
                )
        return existing_stypes

    def _add_datasource_features(
        self,
        datasource: BaseSource,
        ignore_cols: List[str],
        force_stype: MutableMapping[_SemanticTypeValue, List[str]],
        descriptions: Mapping[str, str],
    ) -> None:
        """Add given dataframe to the schema.

        Adds all the features in the dataframe to the schema, using the dtype to decide
        the semantic type of the feature.
        """
        table_dtypes = datasource.get_dtypes(table_name=self.name)
        for item in force_stype.values():
            ignore_cols = _add_this_to_list(item, ignore_cols)
        inferred_semantic_types = self._dtype_based_stype_split(
            table_dtypes, ignore_cols
        )
        semantic_types = self._combine_existing_stypes_with_forced_stypes(
            inferred_semantic_types, force_stype, table_dtypes
        )
        if SemanticType.CATEGORICAL in semantic_types:
            values = datasource.get_values(
                semantic_types[SemanticType.CATEGORICAL], table_name=self.name
            )

        for stype, features in semantic_types.items():
            # Sort the list of features.
            # This ensures they are added in deterministic order.
            features.sort()
            for feature_name in features:
                dtype = table_dtypes[feature_name]
                description = descriptions.get(feature_name)
                if stype == SemanticType.TEXT:
                    if feature_name not in self.get_feature_names():
                        TextRecord.add_record_to_schema(
                            self,
                            feature_name=feature_name,
                            dtype=dtype,
                            description=description,
                        )
                elif stype == SemanticType.CONTINUOUS:
                    if feature_name not in self.get_feature_names():
                        ContinuousRecord.add_record_to_schema(
                            self,
                            feature_name=feature_name,
                            dtype=dtype,
                            description=description,
                        )

                elif stype == SemanticType.CATEGORICAL:
                    # Convert arbitrary iterable to the right form
                    feature_values: Union[np.ndarray, pd.Series]
                    if not isinstance(
                        (raw_feature_values := values[feature_name]),
                        (np.ndarray, pd.Series),
                    ):
                        feature_values = np.asarray(raw_feature_values)
                    else:
                        feature_values = raw_feature_values

                    self._add_categorical_feature(
                        name=feature_name,
                        dtype=dtype,
                        values=feature_values,
                        description=description,
                    )
                elif stype == SemanticType.IMAGE:
                    if (
                        "image" not in self.features
                        or feature_name not in self.features["image"]
                    ):
                        ImageRecord.add_record_to_schema(
                            self,
                            feature_name=feature_name,
                            dtype=dtype,
                            description=descriptions.get(feature_name),
                        )
                    record = self.features["image"][feature_name]
                    for img in datasource.get_column(
                        feature_name, table_name=self.name
                    ):
                        im = Image.open(img)
                        record.dimensions[im.size] += 1
                        record.modes[im.mode] += 1
                        record.formats[im.format] += 1

    def _expand_dataframe(self, dataframe: pd.DataFrame) -> pd.DataFrame:
        """Expands dataframe to include missing columns specified in the schema.

        Simply adds columns populated with default values: 'nan' for categorical
        and text columns and '0' for continuous columns.

        Args:
            dataframe: dataframe without all the required columns

        Returns:
            Dataframe that includes all the required columns

        Raises:
            BitfountSchemaError: if there is a missing image column as this cannot be
                replicated.
        """
        missing_categorical_value = "nan"
        missing_text_value = "nan"
        missing_continuous_value = 0
        columns = list(dataframe.columns)
        for stype in self.features:
            # Iterate through semantic types
            for feature_name in self.features[cast(_SemanticTypeValue, stype)]:
                # Iterate through each feature in given semantic type
                if feature_name not in columns:
                    # If feature is not present in the given dataframe, add that feature
                    # with a dummy value to the dataframe
                    logger.debug(
                        f"Feature present in schema but missing in data: {feature_name}"
                    )
                    if stype == SemanticType.IMAGE.value:
                        raise BitfountSchemaError(
                            f"Missing image feature {feature_name} in dataframe. "
                            "Unable to apply schema to this dataframe"
                        )
                    elif stype == SemanticType.TEXT.value:
                        dataframe[feature_name] = missing_text_value
                    elif stype == SemanticType.CONTINUOUS.value:
                        dataframe[feature_name] = missing_continuous_value
                    elif stype == SemanticType.CATEGORICAL.value:
                        dataframe[feature_name] = missing_categorical_value
                        # adds the missing categorical value (i.e. 'nan') to the encoder
                        # for the missing categorical feature
                        self._add_categorical_feature(
                            name=feature_name,
                            values=np.array([missing_categorical_value]),
                        )
        return dataframe

    def _reduce_dataframe(
        self, dataframe: pd.DataFrame, keep_cols: Optional[List[str]] = None
    ) -> pd.DataFrame:
        """Drops any columns that are not part of the schema.

        Args:
            dataframe: dataframe which includes extra columns
            keep_cols: optional list of columns to keep even if
                they are not part of the schema. Defaults to None.

        Returns:
            Dataframe with extra columns removed
        """
        cols_to_keep = self.get_feature_names()
        cols_to_keep = _add_this_to_list(keep_cols, cols_to_keep)
        return dataframe[cols_to_keep]

    def _apply_types(
        self, dataframe: pd.DataFrame, selected_cols: Optional[List[str]] = None
    ) -> pd.DataFrame:
        """Applies the prescribed feature types to the dataframe.

        Args:
            dataframe: dataframe with varied types
            selected_cols: optional list of columns selected for
                training to which the types will be applied.

        Returns:
            Dataframe with types that are specified in the schema
        """
        if selected_cols:
            selected = [
                feature_name
                for stype in self.features
                for feature_name, record in self.features[
                    cast(_SemanticTypeValue, stype)
                ].items()
                if feature_name in selected_cols
            ]
        else:
            selected = [
                feature_name
                for stype in self.features
                for feature_name, record in self.features[
                    cast(_SemanticTypeValue, stype)
                ].items()
            ]
        types: Dict[str, Union[Dtype, np.dtype]] = {
            feature_name: record.dtype
            for stype in self.features
            for feature_name, record in self.features[
                cast(_SemanticTypeValue, stype)
            ].items()
            if feature_name in selected
        }

        if "categorical" in self.features:
            types.update(
                {
                    feature_name: record.encoder.dtype
                    for feature_name, record in self.features["categorical"].items()
                    if feature_name in selected
                }
            )

        return dataframe.astype(types)

    def _encode_dataframe(self, data: pd.DataFrame) -> pd.DataFrame:
        """Encodes the dataframe categorical columns according to the schema.

        Args:
            data: the dataframe to be encoded

        Raises:
            ValueError: if the encoder fails to encode a particular column

        Returns:
            The dataframe with the categorical columns encoded
        """
        if "categorical" in self.features:
            for feature_name, record in self.features["categorical"].items():
                if feature_name not in data:
                    logger.warning(
                        f"Column {feature_name} is not in the dataframe. "
                        "Skipping encoding"
                    )
                    continue
                try:
                    data[feature_name] = record.encoder.transform(data[feature_name])
                except ValueError as err:
                    raise ValueError(
                        f"Could not encode column {feature_name}: {str(err)}"
                    )
        else:
            logger.info("No encoding to be done as there are no categorical features.")

        return data

    def get_feature_names(
        self, semantic_type: Optional[SemanticType] = None
    ) -> List[str]:
        """Returns the names of all the features in the schema.

        Args:
            semantic_type (SemanticType, optional): if semantic type is provided, only
                the feature names corresponding to the semantic type are returned.
                Defaults to None.

        Returns:
            features: A list of feature names.
        """
        if semantic_type is not None:
            stype = cast(_SemanticTypeValue, semantic_type.value)
            if stype in self.features:
                features = list(self.features[stype])
            else:
                logger.debug(f"There are no features with semantic type {stype}")
                features = []

        else:
            features = [
                feature_name
                for stype in self.features
                for feature_name in self.features[cast(_SemanticTypeValue, stype)]
            ]
        return features

    def get_categorical_feature_size(self, var: Union[str, List[str]]) -> int:
        """Gets the column dimensions.

        Args:
            var: A column name or a list of column names for which
                to get the dimensions.

        Returns:
            The number of unique value in the categorical column.
        """
        if isinstance(var, list):
            var = var[0]

        if "categorical" not in self.features:
            raise ValueError("No categorical features.")
        elif var not in self.features["categorical"]:
            raise ValueError(f"{var} feature not found in categorical features.")
        return self.features["categorical"][var].encoder.size

    def get_categorical_feature_sizes(
        self, ignore_cols: Optional[Union[str, List[str]]] = None
    ) -> List[int]:
        """Returns a list of categorical feature sizes.

        Args:
            ignore_cols: The column(s) to be ignored from the schema.
        """
        if not ignore_cols:
            ignore_cols = []
        elif isinstance(ignore_cols, str):
            ignore_cols = [ignore_cols]
        return [
            self.get_categorical_feature_size(var)
            for var in self.get_feature_names(SemanticType.CATEGORICAL)
            if var not in ignore_cols
        ]

    def get_num_categorical(
        self, ignore_cols: Optional[Union[str, List[str]]] = None
    ) -> int:
        """Get the number of (non-ignored) categorical features.

        Args:
            ignore_cols: Columns to ignore when counting categorical features.

        Return:
            The number of categorical features.
        """
        if not ignore_cols:
            ignore_cols = []
        elif isinstance(ignore_cols, str):
            ignore_cols = [ignore_cols]
        return len(
            [
                None
                for var in self.get_feature_names(SemanticType.CATEGORICAL)
                if var not in ignore_cols
            ]
        )

    def get_num_continuous(
        self, ignore_cols: Optional[Union[str, List[str]]] = None
    ) -> int:
        """Get the number of (non-ignored) continuous features.

        Args:
            ignore_cols: Columns to ignore when counting continuous features.

        Return:
            The number of continuous features.
        """
        if not ignore_cols:
            ignore_cols = []
        elif isinstance(ignore_cols, str):
            ignore_cols = [ignore_cols]
        return len(
            [
                None
                for var in self.get_feature_names(SemanticType.CONTINUOUS)
                if var not in ignore_cols
            ]
        )

    def add_datasource_features(
        self,
        datasource: BaseSource,
        ignore_cols: Optional[Sequence[str]] = None,
        force_stype: Optional[MutableMapping[_SemanticTypeValue, List[str]]] = None,
        descriptions: Optional[Mapping[str, str]] = None,
    ) -> None:
        """Adds datasource features to schema.

        Args:
            datasource: The datasource whose features this method adds.
            ignore_cols: Columns to ignore from the `BaseSource`. Defaults to None.
            force_stype: Columns for which to change the semantic type.
                Format: semantictype: [columnnames]. Defaults to None.
                Example: {'categorical': ['target_column'],
                        'continuous': ['age', 'salary']}
            descriptions: Descriptions of the features. Defaults to None.

        Raises:
            BitfountSchemaError: if the schema is already frozen
        """
        if ignore_cols is None:
            ignore_cols_aux = []
        else:
            ignore_cols_aux = list(ignore_cols)

        if force_stype is None:
            force_stype = {}

        self._add_datasource_features(
            datasource=datasource,
            ignore_cols=ignore_cols_aux,
            force_stype=force_stype,
            descriptions=descriptions if descriptions is not None else {},
        )

    def apply(
        self, dataframe: pd.DataFrame, keep_cols: Optional[List[str]] = None
    ) -> pd.DataFrame:
        """Applies the schema to a dataframe and returns the transformed dataframe.

        Sequentially adds missing columns to the dataframe, removes superfluous columns
        from the dataframe, changes the types of the columns in the dataframe and
        finally encodes the categorical columns in the dataframe before returning the
        transformed dataframe.

        Args:
            dataframe: The dataframe to transform.
            keep_cols: A list of columns to keep even if
                they are not part of the schema. Defaults to None.

        Returns:
            The dataframe with the transformations applied.
        """
        dataframe = self._expand_dataframe(dataframe)
        dataframe = self._reduce_dataframe(dataframe, keep_cols=keep_cols)
        dataframe = self._apply_types(dataframe, selected_cols=keep_cols)
        dataframe = self._encode_dataframe(dataframe)
        return dataframe

    def __eq__(self, other: Any) -> bool:
        """Compare two _TableSchema objects for equality.

        For two schemas to be equal they must have the same set of features,
        including names and types. This does not require them to have come from
        the same data source though (i.e. their hashes might be different).

        Args:
            other: The other object to compare against.

        Returns:
            True if equal, False otherwise.
        """
        # Check if exact same object
        if self is other:
            return True

        # Check comparable types
        if not isinstance(other, TableSchema):
            return False

        # Check table name
        if self.name != other.name:
            return False

        def extract_features_and_types(
            schema: TableSchema,
        ) -> Dict[str, Dict[str, Tuple[Union[Dtype, np.dtype], SemanticType]]]:
            # Extract types from features
            return {
                feature_type: {
                    feature_name: (record.dtype, record.semantic_type)
                    for feature_name, record in cast(
                        Dict[str, _SemanticTypeRecord], records_dict
                    ).items()
                }
                for feature_type, records_dict in schema.features.items()
            }

        # Check features and their types
        if extract_features_and_types(self) != extract_features_and_types(other):
            return False

        # Otherwise, equal for our purposes
        return True


class _BitfountSchemaMarshmallowMixIn:
    """MixIn class for Schema serialization."""

    def dump(self, file_path: PathLike) -> None:
        """Dumps the schema as a yaml file.

        Args:
            file_path: The path where the file should be saved

        Returns:
            none
        """
        with open(file_path, "w") as file:
            file.write(self.dumps())

    def dumps(self) -> Any:
        """Produces the YAML representation of the schema object.

        Returns:
            str: The YAML representation of the schema
        """
        return yaml.dump(self.to_json(), sort_keys=False)

    def to_json(self) -> _JSONDict:
        """Turns a schema object into a JSON compatible dictionary.

        Returns:
            dict: A simple JSON compatible representation of the Schema
        """
        # Our self._Schema() objects are dumped as JSON-compatible dicts
        return cast(_JSONDict, self._Schema().dump(self))

    @classmethod
    def load(cls, data: Mapping) -> BitfountSchema:
        """Loads the schema from a dictionary.

        Args:
            data: The data to load the schema from.

        Returns:
            BitfountSchema: A _Schema instance corresponding to the dictionary.
        """
        # @post_load guarantees this will be a BitfountSchema
        schema: BitfountSchema = cls._Schema().load(data)
        return schema

    @classmethod
    def load_from_file(cls, file_path: Union[str, PathLike]) -> BitfountSchema:
        """Loads the schema from a yaml file.

        This contains validation errors to help fix an invalid YAML file.
        """
        with open(file_path, "r") as f:
            schema_as_yaml = yaml.safe_load(f)
        return cls.load(schema_as_yaml)

    class _Schema(_CamelCaseSchema):
        tables = fields.List(fields.Nested(TableSchema._Schema))
        # TODO: [BIT-1057] Consider moving metadata to be a separate part of the
        #       output YAML.
        # To maintain backwards compatibility with schemas that may not contain
        # metadata we use a default value.
        metadata = fields.Method(
            serialize="dump_metadata", deserialize="load_metadata", load_default=dict
        )

        @staticmethod
        def dump_metadata(obj: BitfountSchema) -> Dict[str, str]:
            """Creates and dumps metadata for the schema."""
            return {
                "bitfount_version": bitfount.__version__,
                "hash": obj.hash,
                "schema_version": "1",
            }

        @staticmethod
        def load_metadata(value: Dict[str, str]) -> Dict[str, str]:
            """Loads the metadata dict."""
            return value

        @post_dump
        def sort_tables(self, data: _JSONDict, **kwargs: Any) -> _JSONDict:
            """Sorts tables alphabetically. by name."""
            new_data = {}
            # Type ignore due to this bug: https://github.com/python/mypy/issues/9656
            new_data["tables"] = sorted(
                data["tables"], key=lambda d: d["name"]  # type: ignore[no-any-return] # Reason: see comment above # noqa: B950
            )
            # Metadata is added as a key after the tables to ensure it ends up at the
            # bottom of the YAML file.
            new_data["metadata"] = data["metadata"]
            return new_data

        @post_load
        def recreate_schema(self, data: _JSONDict, **_kwargs: Any) -> BitfountSchema:
            """Recreates Schema."""
            new_schema = BitfountSchema()
            new_schema.tables = data["tables"]
            # Ensure existing datasources hash is loaded if present
            new_schema._orig_hash = data["metadata"].get("hash")
            return new_schema


@delegates()
class BitfountSchema(_BitfountSchemaMarshmallowMixIn):
    """A schema that defines the tables of a `BaseSource`.

    It lists all the tables found in the `BaseSource` and the features in those tables.

    Args:
        datasource: An optional `BaseSource` object. This can be provided as a utility
            to avoid calling `add_datasource_tables`.
        **kwargs: Optional keyword arguments to be provided to `add_datasource_tables`.
    """

    def __init__(self, datasource: Optional[BaseSource] = None, **kwargs: Any):
        self.tables: List[TableSchema] = []
        # self._orig_hash is used to store the hash when loading a previously
        # generated schema.
        self._orig_hash: Optional[str] = None
        # Datasource hashes is a set to ensure that adding the same datasource multiple
        # times does not result in a different hash.
        self._datasource_hashes: Set[str] = set()
        # Used to stop any more datasources from being added
        self._frozen: bool = False
        if datasource is not None:
            self.add_datasource_tables(datasource, **kwargs)

    @property
    def table_names(self) -> List[str]:
        """Returns a list of table names."""
        return [table.name for table in self.tables]

    @property
    def hash(self) -> str:
        """The hash of this schema.

        This relates to the BaseSource(s) that were used in the generation of this
        schema to assure that this schema is used against compatible data sources.

        Returns:
            A sha256 hash of the `_datasource_hashes`.
        """
        # Must be sorted to ensure ordering of BaseSources being added doesn't
        # change things.
        frozen_hashes: str = str(sorted(self._datasource_hashes))
        return _hash_str(frozen_hashes)

    def freeze(self) -> None:
        """Freezes the schema, ensuring no more datasources can be added.

        If this schema was loaded from an already generated schema, this will
        also check that the schema is compatible with the datasources set.
        """
        if self._orig_hash and self.hash != self._orig_hash:
            raise BitfountSchemaError(
                "This schema was generated against a different set of datasources "
                "and is incompatible with those selected. This may be due to "
                "changing column names or types. Please generate a new schema."
            )
        self._frozen = True

    def unfreeze(self) -> None:
        """Unfreezes the schema, allowing more datasources to be added."""
        self._frozen = False

    def add_datasource_tables(
        self,
        datasource: BaseSource,
        table_name: Optional[str] = None,
        table_descriptions: Optional[Mapping[str, str]] = None,
        column_descriptions: Optional[Mapping[str, Mapping[str, str]]] = None,
        ignore_cols: Optional[Mapping[str, Sequence[str]]] = None,
        force_stypes: Optional[
            Mapping[str, MutableMapping[_SemanticTypeValue, List[str]]]
        ] = None,
    ) -> None:
        """Adds the tables from a `BaseSource` to the schema.

        Args:
            datasource: The `BaseSource` to add the tables from.
            table_name: The name of the table if there is only one table in the
                `BaseSource`.
            table_descriptions: A mapping of table names to descriptions.
            column_descriptions: A mapping of table names to a mapping of column names
                to descriptions.
            ignore_cols: A mapping of table names to a list of column names to ignore.
            force_stypes: A mapping of table names to a mapping of semantic types to
                a list of column names.

        Raises:
            BitfountSchemaError: If the schema is already frozen.
            DataSourceError: If the `BaseSource` for the provided datasource has
                not been initialised properly. This can be done by calling
                `super().__init__(**kwargs)` in the `__init__` of the DataSource.
            ValueError: If a table name hasn't been provided for a single table
                `BaseSource`.
        """
        # Raise error if BaseSource has not been initialised
        if not datasource.is_initialised:
            raise DataSourceError(
                "The datasource provided has not initialised the BaseSource "
                "parent class. Please make sure that you call "
                "`super().__init__(**kwargs)` in your child method."
            )
        if not self._frozen:

            if datasource.multi_table:
                assert isinstance(datasource, MultiTableSource)  # nosec[assert_used]
                assert datasource.table_names is not None  # nosec[assert_used]

                table_names = datasource.table_names
                if table_name:
                    logger.warning(
                        "Ignoring table_name argument for multi-table datasource."
                    )
            elif table_name:
                table_names = [table_name]
            else:
                raise ValueError(
                    "Must provide a table name for single table datasources."
                )

            # Iterate over the tables in the datasource and add them to the schema
            for t in table_names:
                table_description = (
                    None if not table_descriptions else table_descriptions.get(t)
                )
                column_descriptions_ = (
                    None if not column_descriptions else column_descriptions.get(t)
                )
                ignore_cols_ = None if not ignore_cols else ignore_cols.get(t)
                force_stype = None if not force_stypes else force_stypes.get(t)

                # Create table if it doesn't exist
                if t not in self.table_names:
                    self.tables.append(TableSchema(t, table_description))

                table_schema = self.get_table_schema(t)
                table_schema.add_datasource_features(
                    datasource=datasource,
                    descriptions=column_descriptions_,
                    ignore_cols=ignore_cols_,
                    force_stype=force_stype,
                )

            self._datasource_hashes.add(datasource.hash)

        else:
            raise BitfountSchemaError(
                "This schema is frozen. No more datasources can be added."
            )

    def get_table_schema(self, table_name: str) -> TableSchema:
        """Gets a table schema from the schema.

        Args:
            table_name: The name of the table schema to get.

        Returns:
            The table with the given name.

        Raises:
            BitfountSchemaError: If the table is not found.
        """
        for table in self.tables:
            if table.name == table_name:
                return table
        raise BitfountSchemaError(f"Table {table_name} not found in schema.")

    def get_feature_names(
        self, table_name: str, semantic_type: Optional[SemanticType] = None
    ) -> List[str]:
        """Returns the names of all the features in the schema.

        Args:
            table_name: The name of the table to get the features from.
            semantic_type: if semantic type is provided, only
                the feature names corresponding to the semantic type are returned.
                Defaults to None.

        Returns:
            features: A list of feature names.
        """
        table_schema = self.get_table_schema(table_name)
        return table_schema.get_feature_names(semantic_type)

    def get_categorical_feature_size(
        self, table_name: str, var: Union[str, List[str]]
    ) -> int:
        """Gets the column dimensions.

        Args:
            table_name: The name of the table to get the column dimensions from.
            var: A column name or a list of column names for which
                to get the dimensions.

        Returns:
            The number of unique value in the categorical column.
        """
        table_schema = self.get_table_schema(table_name)
        return table_schema.get_categorical_feature_size(var)

    def get_categorical_feature_sizes(
        self, table_name: str, ignore_cols: Optional[Union[str, List[str]]] = None
    ) -> List[int]:
        """Returns a list of categorical feature sizes.

        Args:
            table_name: The name of the table to get the categorical feature sizes.
            ignore_cols: The column(s) to be ignored from the schema.
        """
        table_schema = self.get_table_schema(table_name)
        return table_schema.get_categorical_feature_sizes(ignore_cols)

    def apply(
        self, dataframe: pd.DataFrame, keep_cols: Optional[List[str]] = None
    ) -> pd.DataFrame:
        """Applies the schema to a dataframe and returns the transformed dataframe.

        Sequentially adds missing columns to the dataframe, removes superfluous columns
        from the dataframe, changes the types of the columns in the dataframe and
        finally encodes the categorical columns in the dataframe before returning the
        transformed dataframe.

        Args:
            dataframe: The dataframe to transform.
            keep_cols: A list of columns to keep even if
                they are not part of the schema. Defaults to None.

        Returns:
            The dataframe with the transformations applied.

        Raises:
            BitfountSchemaError: If the schema cannot be applied to the dataframe.
        """
        if len(self.tables) == 0:
            raise BitfountSchemaError("No tables in schema.")
        elif len(self.tables) > 1:
            raise BitfountSchemaError(
                "Can't apply a multi-table schema to a single dataframe."
            )

        table_schema = self.tables[0]
        return table_schema.apply(dataframe, keep_cols)

    def __eq__(self, other: Any) -> bool:
        """Compare two BitfountSchema objects for equality.

        For two schemas to be equal they must have the same set of table names and
        contents. This does not require them to have come from the same data source
        though (i.e. their hashes might be different).

        Args:
            other: The other object to compare against.

        Returns:
            True if equal, False otherwise.
        """
        # Check if exact same object
        if self is other:
            return True

        # Check comparable types
        if not isinstance(other, BitfountSchema):
            return False

        # Check if same number of tables before iterating
        if len(self.tables) != len(other.tables):
            return False

        # Iterate through tables sorted by name and check equality
        for t1, t2 in zip(
            sorted(self.tables, key=lambda d: d.name),
            sorted(other.tables, key=lambda d: d.name),
        ):
            if t1 == t2:
                continue
            return False

        return True



"""Data factory classes.

Factory design patterns for producing datasets and dataloaders.
"""
from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, List, Mapping, Optional, Union

from bitfount.config import _BASIC_ENGINE, _PYTORCH_ENGINE, BITFOUNT_ENGINE
from bitfount.data.dataloaders import BitfountDataLoader
from bitfount.data.datasets import _BaseBitfountDataset, _BitfountDataset
from bitfount.exceptions import BitfountEngineError

if TYPE_CHECKING:
    from bitfount.data.datasources.base_source import BaseSource
    from bitfount.data.datasplitters import DatasetSplitter
    from bitfount.data.schema import TableSchema
    from bitfount.data.types import DataSplit, _SemanticTypeValue
    from bitfount.transformations.batch_operations import BatchTimeOperation


class _DataFactory(ABC):
    """A factory for producing dataset and dataloader instances."""

    def __init__(self, *args: Any, **kwargs: Any):
        pass

    @abstractmethod
    def create_dataloader(
        self, dataset: _BaseBitfountDataset, batch_size: Optional[int] = None
    ) -> BitfountDataLoader:
        """Creates a dataloader as specified by this factory.

        Args:
            dataset: The dataset that should be loaded.
            batch_size: The batch size that the dataloader should output.

        Returns:
            A BitfountDataLoader instance.
        """
        raise NotImplementedError

    @abstractmethod
    def create_dataset(
        self,
        datasource: BaseSource,
        data_splitter: Optional[DatasetSplitter],
        data_split: DataSplit,
        schema: TableSchema,
        selected_cols: List[str],
        selected_cols_semantic_types: Mapping[_SemanticTypeValue, List[str]],
        target: Optional[Union[str, List[str]]] = None,
        batch_transforms: Optional[List[BatchTimeOperation]] = None,
        **kwargs: Any,
    ) -> _BaseBitfountDataset:
        """Creates a dataset for prediction tasks.

        Args:
            datasource: The data to wrap in a dataset.
            data_splitter: The data splitter to use, as specified by the Modeller.
            data_split: Which data split of the datasource to use and the step for which
                batch transformations should be applied.
            schema: The schema of the datasource.
            selected_cols: The columns selected by the data structure.
            selected_cols_semantic_types: A mapping of semantic types and column names.
            target: The dependent variable name.
            batch_transforms: A list of transformations to be applied at batch time.
            **kwargs: Any
        """
        raise NotImplementedError


class _BasicDataFactory(_DataFactory):
    """A basic DataFactory implementation using core dataset and dataloaders."""

    def create_dataloader(
        self, dataset: _BaseBitfountDataset, batch_size: Optional[int] = None
    ) -> BitfountDataLoader:
        """See base class."""
        return BitfountDataLoader(dataset, batch_size)

    def create_dataset(
        self,
        datasource: BaseSource,
        data_splitter: Optional[DatasetSplitter],
        data_split: DataSplit,
        schema: TableSchema,
        selected_cols: List[str],
        selected_cols_semantic_types: Mapping[_SemanticTypeValue, List[str]],
        target: Optional[Union[str, List[str]]] = None,
        batch_transforms: Optional[List[BatchTimeOperation]] = None,
        **kwargs: Any,
    ) -> _BaseBitfountDataset:
        """See base class."""
        return _BitfountDataset(
            datasource=datasource,
            schema=schema,
            data_splitter=data_splitter,
            data_split=data_split,
            target=target,
            selected_cols=selected_cols,
            selected_cols_semantic_types=selected_cols_semantic_types,
            batch_transforms=batch_transforms,
            **kwargs,
        )


def _get_default_data_factory(*args: Any, **kwargs: Any) -> _DataFactory:
    """Create a DataFactory instance as specified by the backend engine.

    Args:
        *args: positional arguments, passed to the DataFactory constructor.
        **kwargs: keyword arguments, passed to the DataFactory constructor.

    Returns:
        The created DataFactory instance.

    Raises:
        BitfountEngineError: if there is an import issue in loading the backend.
    """
    if BITFOUNT_ENGINE == _PYTORCH_ENGINE:
        try:
            from bitfount.backends.pytorch.data.datafactory import _PyTorchDataFactory

            return _PyTorchDataFactory(*args, **kwargs)
        except ImportError:
            raise BitfountEngineError(
                "An error was encountered trying to load the pytorch engine; "
                "check pytorch is installed."
            )
    elif BITFOUNT_ENGINE == _BASIC_ENGINE:
        return _BasicDataFactory(*args, **kwargs)
    else:
        raise BitfountEngineError(f"Unable to load engine {BITFOUNT_ENGINE}.")



"""Dataset-related transformations.

This module contains the base class and concrete classes for dataset transformations,
those that potentially act over the entire dataset.
"""
from typing import List, Mapping, Union

import attr

from bitfount.transformations.base_transformation import Transformation
from bitfount.transformations.exceptions import TransformationApplicationError
from bitfount.utils import delegates


@delegates()
@attr.dataclass(kw_only=True)
class DatasetTransformation(Transformation):
    """Base transformation for all dataset transformation classes.

    User can specify "all" to have it act on every relevant column as defined
    in the schema.

    Args:
        output: Whether or not this transformation should be included in the final
            output. This must be True for all dataset transformations. Defaults to True.
        cols: The columns to act on as a list of strings. Defaults to "all" which acts
            on all columns in the dataset.

    Raises:
        ValueError: If `output` is False.
    """

    output: bool = True
    cols: Union[str, List[str]] = "all"

    def __attrs_post_init__(self) -> None:
        super().__attrs_post_init__()
        if not self.output:
            raise ValueError("`output` cannot be False for a DatasetTransformation")


@delegates()
@attr.dataclass(kw_only=True)
class CleanDataTransformation(DatasetTransformation):
    """Dataset transformation that will "clean" the specified columns.

    For continuous columns this will replace all infinities and NaNs with 0.
    For categorical columns this will replace all NaN's with "nan" explicitly.
    """

    _registry_name = "cleandata"


@delegates()
@attr.dataclass(kw_only=True)
class NormalizeDataTransformation(DatasetTransformation):
    """Dataset transformation that will normalise the specified continuous columns.

    Args:
        cols: The columns to act on as a list of strings. By default, this
            transformation will only apply to columns of type float.

    If this transformation should be applied to all continuous columns,
    the cols attribute should be set to 'all'.
    """

    cols: Union[str, List[str]] = "float"
    _registry_name = "normalizedata"


@delegates()
@attr.dataclass(kw_only=True)
class ScalarMultiplicationDataTransformation(DatasetTransformation):
    """Dataset transformation that multiplies the specified columns by a scalar.

    Transformation applied to the dataset in place. Only applies to continuous columns.

    Args:
        scalar: the scalar to be used for multiplication. It can be provided
            as a number, in which case all numerical columns will be multiplied
            by the respective scalar or as a dictionary mapping column names
            to scalars for multiplication. Defaults to 1.

    Raises:
        TransformationApplicationError: if the scalar variable is not correctly
                instantiated.
    """

    scalar: Union[int, float, Mapping[str, Union[int, float]]] = 1
    _registry_name = "scalarmultiply"

    def __attrs_post_init__(self) -> None:
        if not isinstance(self.scalar, (int, float, dict)):
            raise TransformationApplicationError(
                f"The scalar definition ({self.scalar}) for this transformation is"
                " incorrect. Please make sure you pass an integer, a float, or a"
                " dictionary with column names mapped to the value of the scalar"
                " that the respective column should be mapped to."
            )


@delegates()
@attr.dataclass(kw_only=True)
class ScalarAdditionDataTransformation(DatasetTransformation):
    """Dataset transformation that adds a scalar to the specified columns.

    Transformation applied to the dataset in place. Only applies to continuous columns.

    Args:
        scalar: the scalar to be used for multiplication. It can be provided
            as a number, in which case all numerical columns will be multiplied
            by the respective scalar or as a dictionary mapping column names
            to scalars for multiplication. Defaults to 0.

    Raises:
        TransformationApplicationError: if the scalar variable is not correctly
                instantiated.
    """

    scalar: Union[int, float, Mapping[str, Union[int, float]]] = 0
    _registry_name = "scalaradd"

    def __attrs_post_init__(self) -> None:
        if not isinstance(self.scalar, (int, float, dict)):
            raise TransformationApplicationError(
                f"The scalar definition ({self.scalar}) for this transformation is"
                " incorrect. Please make sure you pass an integer, a float, or a"
                " dictionary with column names mapped to the value of the scalar"
                " that the respective column should be mapped to."
            )



"""Tests for the sql query algorithm."""
from copy import deepcopy
import logging
from typing import TYPE_CHECKING
from unittest.mock import Mock, create_autospec

from _pytest.logging import LogCaptureFixture
import numpy as np
import pandas as pd
from pandas.core.dtypes.common import pandas_dtype
from pytest import fixture
from pytest_mock import MockerFixture
import sqlalchemy

from bitfount.data.datasources.base_source import BaseSource
from bitfount.data.datasources.database_source import DatabaseConnection, DatabaseSource
from bitfount.data.schema import BitfountSchema
from bitfount.data.types import ContinuousRecord
from bitfount.federated.algorithms.base import (
    _BaseAlgorithm,
    _BaseModellerAlgorithm,
    _BaseWorkerAlgorithm,
)
from bitfount.federated.algorithms.private_sql_query import (
    PrivateSqlQuery,
    _ModellerSide,
    _WorkerSide,
)
from bitfount.federated.modeller import _Modeller
from bitfount.federated.privacy.differential import DPPodConfig
from bitfount.federated.utils import _ALGORITHMS
from bitfount.hub import BitfountHub
from bitfount.schemas.utils import bf_dump, bf_load
from tests.utils.helper import (
    TABLE_NAME,
    create_datasource,
    create_schema,
    create_schema_pii,
    integration_test,
    unit_test,
)


@fixture
def column_ranges() -> dict:
    """Fixture for the column ranges."""
    return {
        "A": {"lower": 1, "upper": 1000},  # will be int (A-D)
        "B": {"lower": 1, "upper": 1000},
        "C": {"lower": 1, "upper": 1000},
        "D": {"lower": 1, "upper": 1000},
        "E": {"lower": 0, "upper": 1},  # will be float (E-H)
        "F": {"lower": 0, "upper": 1},
        "G": {"lower": 0, "upper": 1},
        "H": {"lower": 0, "upper": 1},
        "I": {},  # will be string (I-L)
        "J": {},
        "K": {},
        "L": {},
        "TARGET": {"lower": 0, "upper": 1},  # will be int
    }


class TestPrivateSqlQuery:
    """Test PrivateSqlQuery algorithm."""

    @fixture
    def datasource(self) -> BaseSource:
        """Fixture for datasource."""
        return create_datasource(classification=True)

    @fixture
    def column_ranges_multi(self) -> dict:
        """Fixture for the column ranges."""
        return {
            "dummy_data": {
                "name": {"lower": 1, "upper": 5000, "private_id": True},
                "age": {"lower": 1, "upper": 100},
                "weight": {"lower": 1.8, "upper": 230},
                "exercise": {},
            },
            "dummy_data_2": {
                "name": {"lower": 1, "upper": 5000, "private_id": True},
                "height": {"lower": 25, "upper": 230},
            },
        }

    @fixture
    def pod_schema(self) -> BitfountSchema:
        """Fixture for schema."""
        return create_schema(classification=True)

    @fixture
    def pod_schema_pii(self) -> BitfountSchema:
        """Fixture for PII data schema."""
        return create_schema_pii()

    @unit_test
    def test_modeller_types(self, column_ranges: dict) -> None:
        """Test modeller method."""
        algorithm_factory = PrivateSqlQuery(
            query="SELECT * from df.df",
            epsilon=0.1,
            delta=0.00001,
            column_ranges=column_ranges,
        )
        algorithm = algorithm_factory.modeller()
        for type_ in [
            _BaseAlgorithm,
            _BaseModellerAlgorithm,
        ]:
            assert isinstance(algorithm, type_)

    @unit_test
    def test_worker_types(self, column_ranges: dict) -> None:
        """Test worker method."""
        algorithm_factory = PrivateSqlQuery(
            query="SELECT * from df.df",
            epsilon=0.1,
            delta=0.00001,
            column_ranges=column_ranges,
        )
        algorithm = algorithm_factory.worker(
            hub=create_autospec(BitfountHub, instance=True)
        )
        for type_ in [
            _BaseAlgorithm,
            _BaseWorkerAlgorithm,
        ]:
            assert isinstance(algorithm, type_)

    @unit_test
    def test_worker_init_datasource(
        self, column_ranges: dict, datasource: BaseSource
    ) -> None:
        """Test worker init with datasource."""
        kwargs = {"hub": Mock(spec=BitfountHub)}
        algorithm_factory = PrivateSqlQuery(
            query="SELECT * from df.df",
            epsilon=0.1,
            delta=0.00001,
            column_ranges=column_ranges,
            **kwargs,
        )
        algorithm_factory.worker(**kwargs).initialise(
            datasource=datasource,
            pod_identifier="test-pod",
            pod_dp=DPPodConfig(epsilon=1, delta=0.0001),
        )

    @unit_test
    def test_worker_init_missingargs(
        self, column_ranges: dict, datasource: BaseSource
    ) -> None:
        """Test worker init without all arguments."""
        kwargs = {"hub": Mock(spec=BitfountHub)}
        algorithm_factory = PrivateSqlQuery(
            query="SELECT * from df.df",
            epsilon=0.1,
            delta=0.00001,
            column_ranges=column_ranges,
            **kwargs,
        )
        algorithm_factory.worker(**kwargs).initialise(datasource=datasource)

    @unit_test
    def test_modeller_init(self, column_ranges: dict) -> None:
        """Test modeller init method."""
        algorithm_factory = PrivateSqlQuery(
            query="SELECT * from df.df",
            epsilon=0.1,
            delta=0.00001,
            column_ranges=column_ranges,
        )
        algorithm_factory.modeller().initialise()

    @unit_test
    def test_bad_sql_no_table(
        self,
        caplog: LogCaptureFixture,
        column_ranges: dict,
        datasource: BaseSource,
        pod_schema: BitfountSchema,
    ) -> None:
        """Test that having bad SQL query raises an error."""
        # Mock out hub creation
        mock_hub = create_autospec(BitfountHub, instance=True)
        mock_hub.get_pod_schema.return_value = pod_schema
        kwargs = {"hub": mock_hub}

        algorithm_factory = PrivateSqlQuery(
            query="SELECT MAX(G) AS MAX_OF_G",
            epsilon=0.1,
            delta=0.00001,
            column_ranges=column_ranges,
        )
        worker = algorithm_factory.worker(**kwargs)
        worker.datasource = datasource
        worker.pod_identifier = "test-pod"
        worker.pod_dp = DPPodConfig(epsilon=1, delta=0.0001)
        worker.run()
        assert "The default table for single table datasource is the pod" in caplog.text

    @unit_test
    def test_no_pod_identifier(
        self,
        caplog: LogCaptureFixture,
        column_ranges: dict,
        datasource: BaseSource,
        pod_schema: BitfountSchema,
    ) -> None:
        """Test that having no pod identifier raises error."""
        # Mock out hub creation
        mock_hub = create_autospec(BitfountHub, instance=True)
        mock_hub.get_pod_schema.return_value = pod_schema
        kwargs = {"hub": mock_hub}

        algorithm_factory = PrivateSqlQuery(
            query="SELECT AVG(G) AS AVG_OF_G from df.df",
            epsilon=0.1,
            delta=0.00001,
            column_ranges=column_ranges,
        )
        worker = algorithm_factory.worker(**kwargs)
        worker.datasource = datasource
        worker.pod_identifier = None
        worker.pod_dp = DPPodConfig(epsilon=1, delta=0.0001)
        worker.run()
        assert "No pod identifier - cannot get schema" in caplog.text

    @unit_test
    def test_bad_sql_query_statement(
        self,
        caplog: LogCaptureFixture,
        column_ranges: dict,
        datasource: BaseSource,
        pod_schema: BitfountSchema,
    ) -> None:
        """Test that a bad operator in SQL query errors out."""
        # Mock out hub creation
        mock_hub = create_autospec(BitfountHub, instance=True)
        mock_hub.get_pod_schema.return_value = pod_schema
        kwargs = {"hub": mock_hub}
        pod_identifier = "test-pod"

        algorithm_factory = PrivateSqlQuery(
            query="SELECTOR MAX(G) AS MAX_OF_G from `test-pod`.`test-pod`",
            table=pod_identifier,
            epsilon=0.1,
            delta=0.00001,
            column_ranges=column_ranges,
        )
        worker = algorithm_factory.worker(**kwargs)
        worker.datasource = datasource
        worker.pod_identifier = "test-pod"
        worker.pod_dp = DPPodConfig(epsilon=1, delta=0.0001)
        worker.run()
        assert "Error executing PrivateSQL query" in caplog.text

    @unit_test
    def test_bad_sql_query_column(
        self,
        caplog: LogCaptureFixture,
        column_ranges: dict,
        datasource: BaseSource,
        pod_schema: BitfountSchema,
    ) -> None:
        """Test that an invalid column in SQL query errors out."""
        # Mock out hub creation
        mock_hub = create_autospec(BitfountHub, instance=True)
        mock_hub.get_pod_schema.return_value = pod_schema
        kwargs = {"hub": mock_hub}
        pod_identifier = "test-pod"

        algorithm_factory = PrivateSqlQuery(
            query="SELECT MAX(BITFOUNT_TEST) AS MAX_OF_BITFOUNT_TEST from `test-pod`.`test-pod`",  # noqa: B950
            table=pod_identifier,
            epsilon=0.1,
            delta=0.00001,
            column_ranges=column_ranges,
        )
        worker = algorithm_factory.worker(**kwargs)
        worker.datasource = datasource
        worker.pod_identifier = pod_identifier
        worker.pod_dp = DPPodConfig(epsilon=1, delta=0.0001)
        worker.run()
        assert "Error executing PrivateSQL query:" in caplog.text

    @unit_test
    def test_worker_gets_sql_results(
        self, column_ranges: dict, datasource: BaseSource, pod_schema: BitfountSchema
    ) -> None:
        """Test that a SQL query returns correct result."""
        # Mock out hub creation
        mock_hub = create_autospec(BitfountHub, instance=True)
        mock_hub.get_pod_schema.return_value = pod_schema
        kwargs = {"hub": mock_hub}

        algorithm_factory = PrivateSqlQuery(
            query="SELECT AVG(G) AS AVG_OF_G FROM `testpod`.`testpod`",
            table="testpod",
            epsilon=0.1,
            delta=0.00001,
            column_ranges=column_ranges,
        )
        worker = algorithm_factory.worker(**kwargs)
        worker.datasource = datasource
        worker.pod_identifier = "testpod"
        worker.pod_dp = DPPodConfig(epsilon=1, delta=0.0001)
        results = worker.run()

        assert results is not None
        assert len(results) > 0
        # The result should not be the same as the true value
        assert float(results[1][0]) != 0.500129

    @integration_test
    def test_sql_algorithm_db_connection_multitable_onetable(
        self,
        column_ranges_multi: dict,
        db_session_pii: sqlalchemy.engine.base.Engine,
        pod_schema_pii: BitfountSchema,
    ) -> None:
        """Test sql algorithm on multitable db connection, check ranges."""
        pod_identifier = "testpod"
        # Mock out hub creation
        pod_schema_pii.tables.append(deepcopy(pod_schema_pii.tables[0]))
        pod_schema_pii.tables[0].name = "dummy_data"
        pod_schema_pii.tables[1].name = "dummy_data_2"

        mock_hub = create_autospec(BitfountHub, instance=True)
        mock_hub.get_pod_schema.return_value = pod_schema_pii
        kwargs = {"hub": mock_hub}

        db_src = DatabaseSource(
            DatabaseConnection(
                db_session_pii, table_names=["dummy_data", "dummy_data_2"]
            )
        )
        algorithm_factory = PrivateSqlQuery(
            query="""
            SELECT avg(age) as avg_adult
            FROM public.dummy_data
            WHERE age > 17;""",
            table=pod_identifier,
            epsilon=0.1,
            delta=0.00001,
            column_ranges=column_ranges_multi,
        )
        worker = algorithm_factory.worker(**kwargs)
        worker.datasource = db_src
        worker.pod_identifier = pod_identifier
        worker.pod_dp = DPPodConfig(epsilon=1, delta=0.0001)
        results = worker.run()

        # Check output has the correct form
        assert results is not None
        assert len(results) > 0
        assert len(results) == 2

        # Check the results are as expected (name and within range)
        assert results[0][0] == "avg_adult"
        assert float(results[1][0]) <= 100
        assert float(results[1][0]) >= 1

    @integration_test
    def test_sql_algorithm_db_connection_no_dp(
        self,
        column_ranges_multi: dict,
        db_session_pii: sqlalchemy.engine.base.Engine,
        pod_schema_pii: BitfountSchema,
    ) -> None:
        """Test sql algorithm on multitable db connection, no pod dp."""
        pod_identifier = "testpod"
        # Mock out hub creation
        pod_schema_pii.tables.append(deepcopy(pod_schema_pii.tables[0]))
        pod_schema_pii.tables[0].name = "dummy_data"
        pod_schema_pii.tables[1].name = "dummy_data_2"

        mock_hub = create_autospec(BitfountHub, instance=True)
        mock_hub.get_pod_schema.return_value = pod_schema_pii
        kwargs = {"hub": mock_hub}

        db_src = DatabaseSource(
            DatabaseConnection(
                db_session_pii, table_names=["dummy_data", "dummy_data_2"]
            )
        )
        algorithm_factory = PrivateSqlQuery(
            query="SELECT avg(age) as avg_adult FROM public.dummy_data where age > 17;",  # noqa: B950
            epsilon=0.1,
            delta=0.00001,
            column_ranges=column_ranges_multi,
        )
        worker = algorithm_factory.worker(**kwargs)
        worker.datasource = db_src
        worker.pod_identifier = pod_identifier
        worker.pod_dp = None
        worker.run()

        # Check that the pod_dp is set to the DP configuration given by modeller.
        assert worker.pod_dp is not None
        assert worker.pod_dp.epsilon == 0.1
        assert worker.pod_dp.delta == 0.00001

    @integration_test
    def test_sql_algorithm_db_connection_test_float(
        self,
        column_ranges_multi: dict,
        db_session_pii: sqlalchemy.engine.base.Engine,
        pod_schema_pii: BitfountSchema,
    ) -> None:
        """Test sql algorithm on multitable db connection, check float result."""
        pod_identifier = "testpod"
        # Mock out hub creation
        pod_schema_pii.tables.append(deepcopy(pod_schema_pii.tables[0]))
        pod_schema_pii.tables[0].name = "dummy_data"
        pod_schema_pii.tables[1].name = "dummy_data_2"

        # Create column metadata without ranges to test defaults
        column_ranges_multi_nobounds = column_ranges_multi
        column_ranges_multi_nobounds["dummy_data_2"]["height"] = {}

        mock_hub = create_autospec(BitfountHub, instance=True)
        mock_hub.get_pod_schema.return_value = pod_schema_pii
        kwargs = {"hub": mock_hub}
        db_conn = DatabaseConnection(
            db_session_pii, table_names=["dummy_data", "dummy_data_2"]
        )
        ds = DatabaseSource(db_conn)
        algorithm_factory = PrivateSqlQuery(
            query="SELECT avg(height) as avg_height FROM public.dummy_data_2;",
            epsilon=0.1,
            delta=0.00001,
            column_ranges=column_ranges_multi_nobounds,
        )
        worker = algorithm_factory.worker(**kwargs)
        worker.datasource = ds
        worker.pod_identifier = pod_identifier
        worker.pod_dp = DPPodConfig(epsilon=1, delta=0.0001)
        results = worker.run()

        # Check output has the correct form
        assert results is not None
        assert len(results) > 0
        assert len(results) == 2

        # Check the results are as expected (name and within range)
        assert results[0][0] == "avg_height"
        # There is a very small probability these tests will fail
        assert float(results[1][0]) <= 4
        assert float(results[1][0]) >= -4

    @integration_test
    def test_sql_algorithm_db_connection_test_colnotthere(
        self,
        caplog: LogCaptureFixture,
        column_ranges_multi: dict,
        db_session_pii: sqlalchemy.engine.base.Engine,
        pod_schema_pii: BitfountSchema,
    ) -> None:
        """Test fallback for column not in data schema."""
        pod_identifier = "testpod"
        # Mock out hub creation
        pod_schema_pii.tables.append(deepcopy(pod_schema_pii.tables[0]))
        pod_schema_pii.tables[0].name = "dummy_data"
        pod_schema_pii.tables[1].name = "dummy_data_2"

        # Create column metadata without ranges to test defaults
        column_ranges_multi_nobounds = column_ranges_multi
        column_ranges_multi_nobounds["dummy_data_2"]["dynamic"] = {}

        mock_hub = create_autospec(BitfountHub, instance=True)
        mock_hub.get_pod_schema.return_value = pod_schema_pii
        kwargs = {"hub": mock_hub}
        db_conn = DatabaseConnection(
            db_session_pii, table_names=["dummy_data", "dummy_data_2"]
        )
        ds = DatabaseSource(db_conn, seed=420)
        algorithm_factory = PrivateSqlQuery(
            query="SELECT avg(height) as avg_height FROM public.dummy_data_2;",
            epsilon=0.1,
            delta=0.00001,
            column_ranges=column_ranges_multi_nobounds,
        )
        worker = algorithm_factory.worker(**kwargs)
        worker.datasource = ds
        worker.pod_identifier = pod_identifier
        worker.pod_dp = DPPodConfig(epsilon=1, delta=0.0001)
        worker.run()
        assert "proceed assuming it is a string" in caplog.text

    @integration_test
    def test_sql_algorithm_db_connection_test_tabnotthere(
        self,
        caplog: LogCaptureFixture,
        column_ranges_multi: dict,
        db_session_pii: sqlalchemy.engine.base.Engine,
        pod_schema_pii: BitfountSchema,
    ) -> None:
        """Test query fails if table in metadata not in pod data."""
        # Mock out hub creation
        pod_schema_pii.tables.append(deepcopy(pod_schema_pii.tables[0]))
        pod_schema_pii.tables[0].name = "dummy_data"
        pod_schema_pii.tables[1].name = "dummy_data_2"
        pod_identifier = "testpod"
        # Create column metadata without ranges to test defaults
        column_ranges_multi_nobounds = column_ranges_multi
        column_ranges_multi_nobounds["dummy_data_3"] = {"dynamic": {}}

        mock_hub = create_autospec(BitfountHub, instance=True)
        mock_hub.get_pod_schema.return_value = pod_schema_pii
        kwargs = {"hub": mock_hub}
        db_conn = DatabaseConnection(
            db_session_pii, table_names=["dummy_data", "dummy_data_2"]
        )
        ds = DatabaseSource(db_conn, seed=420)
        algorithm_factory = PrivateSqlQuery(
            query="SELECT avg(height) as avg_height FROM public.dummy_data_2;",
            epsilon=0.1,
            delta=0.00001,
            column_ranges=column_ranges_multi_nobounds,
        )
        worker = algorithm_factory.worker(**kwargs)
        worker.datasource = ds
        worker.pod_identifier = pod_identifier
        worker.pod_dp = DPPodConfig(epsilon=1, delta=0.0001)
        worker.run()
        assert "SQL query will probably fail" in caplog.text

    @integration_test
    def test_sql_algorithm_db_connection_test_incompletemeta(
        self,
        caplog: LogCaptureFixture,
        column_ranges_multi: dict,
        db_session_pii: sqlalchemy.engine.base.Engine,
        pod_schema_pii: BitfountSchema,
    ) -> None:
        """Test the code which auto-completes the metadata if not complete."""
        # Mock out hub creation
        pod_schema_pii.tables.append(deepcopy(pod_schema_pii.tables[0]))
        pod_schema_pii.tables[0].name = "dummy_data"
        pod_schema_pii.tables[1].name = "dummy_data_2"
        pod_identifier = "testpod"
        # Create column metadata without ranges to test defaults
        column_ranges_missing_feature = column_ranges_multi
        del column_ranges_missing_feature["dummy_data"]["weight"]

        mock_hub = create_autospec(BitfountHub, instance=True)
        mock_hub.get_pod_schema.return_value = pod_schema_pii
        kwargs = {"hub": mock_hub}
        db_conn = DatabaseConnection(
            db_session_pii, table_names=["dummy_data", "dummy_data_2"]
        )
        ds = DatabaseSource(db_conn, seed=420)
        ds.load_data()
        algorithm_factory = PrivateSqlQuery(
            query="SELECT avg(height) as avg_height FROM public.dummy_data_2;",
            epsilon=0.1,
            delta=0.00001,
            column_ranges=column_ranges_missing_feature,
        )
        worker = algorithm_factory.worker(**kwargs)
        worker.datasource = ds
        worker.pod_identifier = pod_identifier
        worker.pod_dp = DPPodConfig(epsilon=1, delta=0.0001)
        worker.run()
        assert "lower and upper bound of (0,1) for field [weight]" in caplog.text

    @integration_test
    def test_sql_algorithm_db_connection_test_othertypes(
        self,
        caplog: LogCaptureFixture,
        column_ranges_multi: dict,
        db_session_pii: sqlalchemy.engine.base.Engine,
        pod_schema_pii: BitfountSchema,
    ) -> None:
        """Test metadata auto-complete.

        If a column/feature is missing, test that defaults ranges
        are correctly applied.
        """
        # Mock out hub creation
        pod_schema_pii.tables.append(deepcopy(pod_schema_pii.tables[0]))
        pod_schema_pii.tables[0].name = "dummy_data"
        pod_schema_pii.tables[1].name = "dummy_data_2"

        # Create column metadata without ranges to test defaults
        column_ranges_missing_feature = column_ranges_multi
        del column_ranges_missing_feature["dummy_data"]["weight"]

        mock_hub = create_autospec(BitfountHub, instance=True)
        mock_hub.get_pod_schema.return_value = pod_schema_pii
        kwargs = {"hub": mock_hub}
        db_conn = DatabaseConnection(
            db_session_pii, table_names=["dummy_data", "dummy_data_2"]
        )
        ds = DatabaseSource(db_conn, seed=420)
        algorithm_factory = PrivateSqlQuery(
            query="SELECT avg(height) as avg_height FROM public.dummy_data_2;",
            epsilon=0.1,
            delta=0.00001,
            column_ranges=column_ranges_missing_feature,
        )
        worker = algorithm_factory.worker(**kwargs)
        worker.datasource = ds
        worker.pod_identifier = "testpod"
        worker.pod_dp = DPPodConfig(epsilon=1, delta=0.0001)
        worker.run()
        assert "lower and upper bound of (0,1) for field [weight]" in caplog.text

    @integration_test
    def test_sql_algorithm_db_connection_multitable_groupby(
        self,
        column_ranges_multi: dict,
        db_session_pii: sqlalchemy.engine.base.Engine,
        pod_schema_pii: BitfountSchema,
    ) -> None:
        """Test Group By statement in SQL query ."""
        pod_identifier = "testpod"
        # Mock out hub creation
        pod_schema_pii.tables.append(deepcopy(pod_schema_pii.tables[0]))
        pod_schema_pii.tables[0].name = "dummy_data"
        pod_schema_pii.tables[1].name = "dummy_data_2"

        mock_hub = create_autospec(BitfountHub, instance=True)
        mock_hub.get_pod_schema.return_value = pod_schema_pii
        kwargs = {"hub": mock_hub}
        db_conn = DatabaseConnection(
            db_session_pii, table_names=["dummy_data", "dummy_data_2"]
        )
        ds = DatabaseSource(db_conn, seed=420)
        algorithm_factory = PrivateSqlQuery(
            query=(
                "SELECT avg(age) as avg_adult, exercise as e "
                "FROM public.dummy_data where age > 17 "
                "GROUP BY exercise "
                "ORDER BY avg_adult ASC;"
            ),
            epsilon=0.1,
            delta=0.00001,
            column_ranges=column_ranges_multi,
        )
        worker = algorithm_factory.worker(**kwargs)
        worker.datasource = ds
        worker.pod_identifier = pod_identifier
        worker.pod_dp = DPPodConfig(epsilon=1, delta=0.0001)
        results = worker.run()
        # Check output has the correct form
        assert results is not None
        assert len(results) > 0
        assert len(results) == 12

        # Check the results are as expected (name and within range)
        assert results[0][0] == "avg_adult"
        assert results[0][1] == "e"
        assert float(results[1][0]) <= 100
        for i in range(1, 12):
            assert float(results[i][0]) >= 1
            assert float(results[i][0]) <= 100

    @integration_test
    def test_sql_algorithm_db_connection_multitable_anothertable(
        self,
        column_ranges_multi: dict,
        db_session_pii: sqlalchemy.engine.base.Engine,
        pod_schema_pii: BitfountSchema,
    ) -> None:
        """Test result of SQL algorithm in different table."""
        # Mock out hub creation
        pod_schema_pii.tables.append(deepcopy(pod_schema_pii.tables[0]))
        pod_schema_pii.tables[0].name = "dummy_data"
        pod_schema_pii.tables[1].name = "dummy_data_2"

        mock_hub = create_autospec(BitfountHub, instance=True)
        mock_hub.get_pod_schema.return_value = pod_schema_pii
        kwargs = {"hub": mock_hub}
        db_conn = DatabaseConnection(
            db_session_pii, table_names=["dummy_data", "dummy_data_2"]
        )
        ds = DatabaseSource(db_conn, seed=420)
        algorithm_factory = PrivateSqlQuery(
            query="SELECT AVG(height) AS avg_height FROM public.dummy_data_2;",
            epsilon=0.1,
            delta=0.00001,
            column_ranges=column_ranges_multi,
        )
        worker = algorithm_factory.worker(**kwargs)
        worker.datasource = ds
        worker.pod_identifier = "testpod"
        worker.pod_dp = DPPodConfig(epsilon=1, delta=0.0001)
        results = worker.run()

        # Check output has the correct form
        assert results is not None
        assert len(results) > 0
        assert len(results) == 2

        # Check the results are as expected (name and within range)
        assert results[0][0] == "avg_height"
        assert float(results[1][0]) <= 230
        assert float(results[1][0]) >= 25

    @integration_test
    def test_join_sql_query_statement(
        self,
        caplog: LogCaptureFixture,
        column_ranges_multi: dict,
        db_session_pii: sqlalchemy.engine.base.Engine,
        pod_schema_pii: BitfountSchema,
    ) -> None:
        """Test that a JOIN yields an appropriate error."""
        # Mock out hub creation
        pod_schema_pii.tables.append(deepcopy(pod_schema_pii.tables[0]))
        pod_schema_pii.tables[0].name = "dummy_data"
        pod_schema_pii.tables[1].name = "dummy_data_2"

        mock_hub = create_autospec(BitfountHub, instance=True)
        mock_hub.get_pod_schema.return_value = pod_schema_pii
        kwargs = {"hub": mock_hub}
        db_conn = DatabaseConnection(
            db_session_pii, table_names=["dummy_data", "dummy_data_2"]
        )
        ds = DatabaseSource(db_conn, seed=420)
        algorithm_factory = PrivateSqlQuery(
            query="""
            SELECT 'd1.Date', 'A'
            FROM public.dummy_data d1
            LEFT JOIN public.dummy_data_2 d2
            ON 'd1.Date' = 'd2.Date'
            """,
            epsilon=0.1,
            delta=0.00001,
            column_ranges=column_ranges_multi,
        )
        worker = algorithm_factory.worker(**kwargs)
        worker.datasource = ds
        worker.pod_identifier = ""
        worker.pod_dp = DPPodConfig(epsilon=1, delta=0.0001)
        worker.run()

        assert "JOIN clauses are not supported" in caplog.text

    @integration_test
    def test_sql_algorithm_db_connection_multitable_epsilonspent(
        self,
        caplog: LogCaptureFixture,
        column_ranges_multi: dict,
        db_session_pii: sqlalchemy.engine.base.Engine,
        pod_schema_pii: BitfountSchema,
    ) -> None:
        """Test that epsilon is respected with SQL query."""
        caplog.set_level(logging.INFO)
        # Mock out hub creation
        pod_schema_pii.tables.append(deepcopy(pod_schema_pii.tables[0]))
        pod_schema_pii.tables[0].name = "dummy_data"
        pod_schema_pii.tables[1].name = "dummy_data_2"
        mock_hub = create_autospec(BitfountHub, instance=True)
        mock_hub.get_pod_schema.return_value = pod_schema_pii
        kwargs = {"hub": mock_hub}
        db_conn = DatabaseConnection(
            db_session_pii, table_names=["dummy_data", "dummy_data_2"]
        )

        ds = DatabaseSource(db_conn, seed=420)
        algorithm_factory = PrivateSqlQuery(
            query="SELECT AVG(height) AS avg_height FROM public.dummy_data_2;",
            epsilon=0.4,
            delta=0.00001,
            column_ranges=column_ranges_multi,
        )
        worker = algorithm_factory.worker(**kwargs)
        worker.datasource = ds
        worker.pod_identifier = "test-pod"
        worker.pod_dp = DPPodConfig(epsilon=1, delta=0.0001)
        worker.run()
        # Check that we surface an error for exceeding budget
        assert (
            "Total privacy cost for query will be "
            "(epsilon=0.9999999999166668, delta=1.4999949999983109e-05)" in caplog.text
        )
        assert (
            "Executing SQL query with epsilon 0.3333333333055556 and delta "
            "1e-05 applied on each queried column." in caplog.text
        )

    @unit_test
    def test_sql_algorithm_db_connection_schema_name(
        self,
        caplog: LogCaptureFixture,
        column_ranges_multi: dict,
        db_session_pii: sqlalchemy.engine.base.Engine,
        pod_schema_pii: BitfountSchema,
    ) -> None:
        """Test that the schema name is added in the run method."""
        caplog.set_level(logging.INFO)
        # Mock out hub creation
        pod_schema_pii.tables.append(deepcopy(pod_schema_pii.tables[0]))
        pod_schema_pii.tables[0].name = "dummy_data"
        pod_schema_pii.tables[1].name = "dummy_data_2"
        mock_hub = create_autospec(BitfountHub, instance=True)
        mock_hub.get_pod_schema.return_value = pod_schema_pii
        kwargs = {"hub": mock_hub}
        db_conn = DatabaseConnection(
            db_session_pii, table_names=["dummy_data", "dummy_data_2"]
        )

        ds = DatabaseSource(db_conn, seed=420)
        algorithm_factory = PrivateSqlQuery(
            query="SELECT AVG(height) AS avg_height FROM public.dummy_data_2;",
            epsilon=20,
            delta=0.00001,
            column_ranges=column_ranges_multi,
        )
        worker = algorithm_factory.worker(**kwargs)
        worker.datasource = ds
        worker.pod_identifier = "test-pod"
        worker.run()
        assert worker.db_schema == "public"

    @unit_test
    def test_sql_algorithm_db_connection_wrong_schema_name(
        self,
        caplog: LogCaptureFixture,
        column_ranges_multi: dict,
        db_session_pii: sqlalchemy.engine.base.Engine,
        pod_schema_pii: BitfountSchema,
    ) -> None:
        """Test that wrong schema name raises error."""
        caplog.set_level(logging.INFO)
        # Mock out hub creation
        pod_schema_pii.tables.append(deepcopy(pod_schema_pii.tables[0]))
        pod_schema_pii.tables[0].name = "dummy_data"
        pod_schema_pii.tables[1].name = "dummy_data_2"
        mock_hub = create_autospec(BitfountHub, instance=True)
        mock_hub.get_pod_schema.return_value = pod_schema_pii
        kwargs = {"hub": mock_hub}
        db_conn = DatabaseConnection(
            db_session_pii, table_names=["dummy_data", "dummy_data_2"]
        )

        ds = DatabaseSource(db_conn, seed=420)
        algorithm_factory = PrivateSqlQuery(
            query="SELECT AVG(height) AS avg_height FROM public.dummy_data_2;",
            epsilon=20,
            delta=0.00001,
            column_ranges=column_ranges_multi,
            db_schema="wrong",
        )
        worker = algorithm_factory.worker(**kwargs)
        worker.datasource = ds
        worker.pod_identifier = "test-pod"
        worker.run()
        assert "Schema 'wrong' not found in database." in caplog.text

    @unit_test
    def test_schema_mapping_unknown_column(
        self,
        caplog: LogCaptureFixture,
        column_ranges: dict,
        datasource: BaseSource,
        pod_schema: BitfountSchema,
    ) -> None:
        """Test failure if feature specified not in pod data."""
        caplog.set_level(logging.INFO)
        pod_identifier = "testpod"
        # Mock out hub creation
        mock_hub = create_autospec(BitfountHub, instance=True)
        mock_hub.get_pod_schema.return_value = pod_schema
        kwargs = {"hub": mock_hub}

        # Update column ranges to have a column not in the schema
        column_ranges_bad = column_ranges
        column_ranges_bad["Z"] = {"lower": 0, "upper": 1}

        # Run the query
        algorithm_factory = PrivateSqlQuery(
            query="SELECT AVG(G) AS AVG_OF_G FROM `testpod`.`testpod`",
            table=pod_identifier,
            epsilon=0.1,
            delta=0.00001,
            column_ranges=column_ranges_bad,
        )
        worker = algorithm_factory.worker(**kwargs)
        worker.datasource = datasource
        worker.pod_identifier = pod_identifier
        worker.pod_dp = DPPodConfig(epsilon=1, delta=0.0001)

        worker.run()

        # Check that both the mapping check, and SQL reader raise errors
        assert "No field named 'Z' present in the schema" in caplog.text
        assert "got error ['Z']" in caplog.text

    @unit_test
    def test_schema_mapping_no_bounds(
        self,
        caplog: LogCaptureFixture,
        column_ranges: dict,
        datasource: BaseSource,
        pod_schema: BitfountSchema,
    ) -> None:
        """Test backing off to default bound values."""
        caplog.set_level(logging.INFO)
        pod_identifier = "testpod"
        # Mock out hub creation
        mock_hub = create_autospec(BitfountHub, instance=True)
        mock_hub.get_pod_schema.return_value = pod_schema
        kwargs = {"hub": mock_hub}

        # Update column ranges to have a column not in the schema
        column_ranges_bad = column_ranges
        column_ranges_bad["A"] = {}

        # Run the query
        algorithm_factory = PrivateSqlQuery(
            query="SELECT AVG(G) AS AVG_OF_G FROM `testpod`.`testpod`",
            table=pod_identifier,
            epsilon=0.1,
            delta=0.00001,
            column_ranges=column_ranges_bad,
        )
        worker = algorithm_factory.worker(**kwargs)
        worker.datasource = datasource
        worker.pod_identifier = pod_identifier
        worker.pod_dp = DPPodConfig(epsilon=1, delta=0.0001)
        worker.run()

        assert "Using default lower and upper bound" in caplog.text

    @unit_test
    def test_schema_mapping_unknown_type(
        self,
        caplog: LogCaptureFixture,
        column_ranges: dict,
        datasource: BaseSource,
        pod_schema: BitfountSchema,
    ) -> None:
        """Test schema with unsupported type raises error."""
        caplog.set_level(logging.INFO)
        pod_identifier = "testpod"
        # Unsupported schema
        pod_schema.get_table_schema(TABLE_NAME).features["continuous"][
            "G"
        ] = ContinuousRecord("G", pandas_dtype("datetime64"))

        # Mock out hub creation
        mock_hub = create_autospec(BitfountHub, instance=True)
        mock_hub.get_pod_schema.return_value = pod_schema
        kwargs = {"hub": mock_hub}

        # Run the query
        algorithm_factory = PrivateSqlQuery(
            query="SELECT AVG(G) AS AVG_OF_G FROM `testpod`.`testpod`",
            table=pod_identifier,
            epsilon=0.1,
            delta=0.00001,
            column_ranges=column_ranges,
        )
        worker = algorithm_factory.worker(**kwargs)
        worker.datasource = datasource
        worker.pod_identifier = "testpod"
        worker.pod_dp = DPPodConfig(epsilon=1, delta=0.0001)

        worker.run()

        # Check that both the mapping check, and SQL reader raise errors
        assert "Type datetime64 for column 'G' is not supported" in caplog.text
        assert "must be over numeric or boolean, got string in AVG" in caplog.text

    @unit_test
    def test_modeller_gets_sql_results(
        self, caplog: LogCaptureFixture, column_ranges: dict, pod_schema: BitfountSchema
    ) -> None:
        """Test SQL query returns a result to modeller."""
        caplog.set_level(logging.INFO)
        # Mock out hub creation
        mock_hub = create_autospec(BitfountHub, instance=True)
        mock_hub.get_pod_schema.return_value = pod_schema

        algorithm_factory = PrivateSqlQuery(
            query="SELECT AVG(G) AS AVG_OF_G FROM testpod.testpod",
            table="testpod",
            epsilon=0.1,
            delta=0.001,
            column_ranges=column_ranges,
        )
        modeller = algorithm_factory.modeller()
        data = {"AVG_OF_G": [0.500129]}
        results = {
            "pod1": pd.DataFrame(data),
        }
        returned_results = modeller.run(results=results)
        np.testing.assert_almost_equal(
            returned_results["pod1"]["AVG_OF_G"].to_numpy(),
            results["pod1"]["AVG_OF_G"].to_numpy(),
            decimal=4,
        )

    @unit_test
    def test_different_privacy_different_results(
        self, column_ranges: dict, datasource: BaseSource, pod_schema: BitfountSchema
    ) -> None:
        """Test different DP levels provide different results."""
        pod_identifier = "testpod"
        # Mock out hub creation
        mock_hub = create_autospec(BitfountHub, instance=True)
        mock_hub.get_pod_schema.return_value = pod_schema
        kwargs = {"hub": mock_hub}

        # Run query with good privacy (i.e. low epsilon and delta)
        algorithm_factory = PrivateSqlQuery(
            query="SELECT AVG(G) AS AVG_OF_G FROM `testpod`.`testpod`",
            table=pod_identifier,
            epsilon=0.1,
            delta=0.00001,
            column_ranges=column_ranges,
        )
        worker = algorithm_factory.worker(**kwargs)
        worker.datasource = datasource
        worker.pod_identifier = pod_identifier
        worker.pod_dp = DPPodConfig(epsilon=2, delta=0.001)

        # Run query 10 times, aggregate deviations from truth value
        all_deviations = []
        ground_truth = 0.500129
        for _ in range(10):
            priv_budget, results_good_privacy = worker.run()
            all_deviations.append(np.abs(ground_truth - float(results_good_privacy[0])))
        mean_deviation_good = np.mean(np.array(all_deviations))

        # Run query with bad privacy (i.e. high epsilon and delta)
        algorithm_factory_bad = PrivateSqlQuery(
            query="SELECT AVG(G) AS AVG_OF_G FROM `testpod`.`testpod`",
            table=pod_identifier,
            epsilon=20.0,
            delta=1.0,
            column_ranges=column_ranges,
        )
        worker = algorithm_factory_bad.worker(**kwargs)
        worker.datasource = datasource
        worker.pod_identifier = pod_identifier
        worker.pod_dp = DPPodConfig(epsilon=40, delta=2.0)

        # Run query 10 times, aggregate deviations from truth value
        all_deviations_bad = []
        for _ in range(10):
            priv_budget, results_bad_privacy = worker.run()
            all_deviations_bad.append(
                np.abs(ground_truth - float(results_bad_privacy[0]))
            )
        mean_deviation_bad = np.mean(np.array(all_deviations_bad))

        # Check that deviation is higher with more DP noise
        assert mean_deviation_good > mean_deviation_bad

    @unit_test
    def test_different_privacy_applied(
        self, column_ranges: dict, datasource: BaseSource, pod_schema: BitfountSchema
    ) -> None:
        """Test that DP yields different results over multiple queries."""
        pod_identifier = "testpod"
        # Mock out hub creation
        mock_hub = create_autospec(BitfountHub, instance=True)
        mock_hub.get_pod_schema.return_value = pod_schema
        kwargs = {"hub": mock_hub}

        # Run query with good privacy (i.e. low epsilon and delta)
        algorithm_factory = PrivateSqlQuery(
            query="SELECT AVG(G) AS AVG_OF_G FROM `testpod`.`testpod`",
            table=pod_identifier,
            epsilon=0.1,
            delta=0.00001,
            column_ranges=column_ranges,
        )
        worker = algorithm_factory.worker(**kwargs)
        worker.datasource = datasource
        worker.pod_identifier = pod_identifier
        worker.pod_dp = DPPodConfig(epsilon=1, delta=0.0001)

        # Run query 10 times, ensure results are unique
        all_results = {}
        for _ in range(10):
            priv_budget, results = worker.run()
            result_value = float(results[0])
            assert result_value not in all_results
            all_results[result_value] = 1

    @unit_test
    def test_no_pod_dp(
        self,
        caplog: LogCaptureFixture,
        column_ranges: dict,
        datasource: BaseSource,
        pod_schema: BitfountSchema,
    ) -> None:
        """Test modeller DP is applied to pod with no DP."""
        pod_identifier = "testpod"
        caplog.set_level(logging.INFO)
        # Mock out hub creation
        mock_hub = create_autospec(BitfountHub, instance=True)
        mock_hub.get_pod_schema.return_value = pod_schema
        kwargs = {"hub": mock_hub}

        # Run query with good privacy (i.e. low epsilon and delta)
        algorithm_factory = PrivateSqlQuery(
            query=f"SELECT SUM(G) AS SUM_OF_G FROM {pod_identifier}.{pod_identifier}",
            epsilon=3.0,
            delta=0.1,
            column_ranges=column_ranges,
        )
        worker = algorithm_factory.worker(**kwargs)
        worker.datasource = datasource
        worker.pod_identifier = pod_identifier
        worker.pod_dp = None
        worker.run()

        # Check that the pod_dp is set to the DP configuration given by modeller.
        assert worker.pod_dp is not None
        assert worker.pod_dp.epsilon == 3.0
        assert worker.pod_dp.delta == 0.1

    @unit_test
    def test_pod_dp_prevails(
        self,
        caplog: LogCaptureFixture,
        column_ranges: dict,
        datasource: BaseSource,
        pod_schema: BitfountSchema,
    ) -> None:
        """Test Pod DP is preferred over modeller DP."""
        caplog.set_level(logging.INFO)
        pod_identifier = "testpod"
        # Mock out hub creation
        mock_hub = create_autospec(BitfountHub, instance=True)
        mock_hub.get_pod_schema.return_value = pod_schema
        kwargs = {"hub": mock_hub}

        # Run query with good privacy (i.e. low epsilon and delta)
        algorithm_factory = PrivateSqlQuery(
            query="SELECT SUM(G) AS SUM_OF_G FROM `testpod`.`testpod`",
            table=pod_identifier,
            epsilon=3.0,
            delta=0.1,
            column_ranges=column_ranges,
        )
        worker = algorithm_factory.worker(**kwargs)
        worker.datasource = datasource
        worker.pod_identifier = pod_identifier
        worker.pod_dp = DPPodConfig(epsilon=2.0, delta=0.001)

        worker.run()
        assert "Requested DP max epsilon (3.0) exceeds maximum" in caplog.text
        assert "Requested DP target delta (0.1) exceeds maximum" in caplog.text
        assert (
            "Total privacy cost for query will be "
            "(epsilon=2.0, delta=0.0009997777777778216)" in caplog.text
        )
        assert (
            "Executing SQL query with epsilon 1.0 and "
            "delta 0.0006666666666666666 applied on each queried column." in caplog.text
        )

    @unit_test
    def test_private_sql_execute(
        self, column_ranges: dict, mock_bitfount_session: Mock, mocker: MockerFixture
    ) -> None:
        """Test execute syntactic sugar."""
        query = PrivateSqlQuery(
            query="SELECT * FROM df.df",
            epsilon=0.1,
            delta=0.00001,
            column_ranges=column_ranges,
        )
        pod_identifiers = ["username/pod-id"]

        mock_modeller_run_method = mocker.patch.object(_Modeller, "run")
        query.execute(pod_identifiers=pod_identifiers)
        mock_modeller_run_method.assert_called_once_with(
            pod_identifiers=pod_identifiers, require_all_pods=False
        )


@unit_test
class TestMarshmallowSerialization:
    """Test Marshmallow Serialization for private sql algorithm."""

    def test_serialization(self, column_ranges: dict) -> None:
        """Test Marshmallow Serialization for private sql algorithm."""
        algorithm_factory = PrivateSqlQuery(
            query="SELECT * FROM df.df",
            epsilon=0.1,
            delta=0.00001,
            column_ranges=column_ranges,
        )
        dumped = bf_dump(algorithm_factory)
        loaded = bf_load(dumped, _ALGORITHMS)
        assert isinstance(loaded, PrivateSqlQuery)
        assert algorithm_factory.__dict__ == loaded.__dict__


# Static tests for algorithm-protocol compatibility
if TYPE_CHECKING:
    from typing import cast

    from bitfount.federated.protocols.results_only import (
        _ResultsOnlyCompatibleAlgoFactory_,
        _ResultsOnlyCompatibleModellerAlgorithm,
        _ResultsOnlyDataIncompatibleWorkerAlgorithm,
    )

    # Check compatible with ResultsOnly
    _algo_factory: _ResultsOnlyCompatibleAlgoFactory_ = PrivateSqlQuery(
        query=cast(str, object()),
        epsilon=cast(float, object()),
        delta=cast(float, object()),
        column_ranges=cast(dict, object()),
    )
    _modeller_side: _ResultsOnlyCompatibleModellerAlgorithm = _ModellerSide()
    _worker_side: _ResultsOnlyDataIncompatibleWorkerAlgorithm = _WorkerSide(
        query=cast(str, object()),
        epsilon=cast(float, object()),
        delta=cast(float, object()),
        column_ranges=cast(dict, object()),
        hub=cast(BitfountHub, object()),
        table=cast(str, object()),
    )



"""Tests for config.py."""
import importlib
import sys
from typing import Generator, List, Optional, Tuple

import GPUtil
from _pytest.monkeypatch import MonkeyPatch
import pytest
from pytest import fixture
from pytest_mock import MockerFixture

import bitfount.config
from bitfount.config import (
    _DEVELOPMENT_ENVIRONMENT,
    _PRODUCTION_ENVIRONMENT,
    _STAGING_ENVIRONMENT,
    _get_environment,
)
from tests.utils.helper import unit_test


@unit_test
class TestConfigLoading:
    """Tests for loading variables during config.py import."""

    @fixture
    def config_unimport(self) -> Generator[None, None, None]:
        """Ensures the config module can be freshly imported in the test."""
        # Retrieve and store the old reference to config
        old_ref = sys.modules.pop("bitfount.config")

        yield

        # Restore sys.modules mapping and local namespace
        sys.modules["bitfount.config"] = old_ref
        bitfount.config = old_ref

    def test_config_loading_no_env_var(
        self, config_unimport: None, monkeypatch: MonkeyPatch
    ) -> None:
        """Tests default value assignment if no envvar."""
        monkeypatch.delenv("BITFOUNT_ENGINE", raising=False)
        config = importlib.import_module("bitfount.config")
        assert config.BITFOUNT_ENGINE == "pytorch"

    def test_config_loading_invalid_envvar(
        self, config_unimport: None, monkeypatch: MonkeyPatch
    ) -> None:
        """Tests exception raised if invalid envvar for BITFOUNT_ENGINE."""
        monkeypatch.setenv("BITFOUNT_ENGINE", "not_a_backend")
        with pytest.raises(ValueError, match=".*(not_a_backend).*"):
            importlib.import_module("bitfount.config")

    def test_config_loading_valid_envvar(
        self, config_unimport: None, monkeypatch: MonkeyPatch
    ) -> None:
        """Tests BITFOUNT_ENGINE setting from environment with valid value."""
        monkeypatch.setenv("BITFOUNT_ENGINE", bitfount.config._PYTORCH_ENGINE)
        config = importlib.import_module("bitfount.config")
        assert config.BITFOUNT_ENGINE == bitfount.config._PYTORCH_ENGINE


@unit_test
class TestGetGPUMetadata:
    """Tests retrieving GPU Metadata."""

    @pytest.mark.parametrize(
        "gpu_info,expected_result",
        [
            ([], (None, 0)),  # getGPUs returns empty list
            (
                [
                    GPUtil.GPU(
                        x,
                        f"uuid-{x}",
                        0.5,
                        5.3,
                        0.2,
                        5.0,
                        "NVIDIA xyz",
                        "BitfountGPU",
                        f"123{x}",
                        "None",
                        False,
                        36,
                    )
                    for x in range(3)
                ],
                ("BitfountGPU", 3),
            ),
        ],
    )
    def test_get_gpu_metadata_with_basic_engine(
        self,
        expected_result: Tuple[Optional[str], int],
        gpu_info: List[GPUtil.GPU],
        mocker: MockerFixture,
        monkeypatch: MonkeyPatch,
    ) -> None:
        """Test that GPUUtil can is used to retrieve GPU info."""
        # Set envvar value
        monkeypatch.setattr(
            "bitfount.config.BITFOUNT_ENGINE", bitfount.config._BASIC_ENGINE
        )

        getGPUs = mocker.patch("bitfount.config.GPUtil.getGPUs")
        getGPUs.return_value = gpu_info

        gpu_metadata = bitfount.config.get_gpu_metadata()

        assert gpu_metadata == expected_result

    def test_get_gpu_metadata_with_basic_engine_throws_exception(
        self, mocker: MockerFixture, monkeypatch: MonkeyPatch
    ) -> None:
        """Test that GPUUtil is used to retrieve GPU info."""
        # Set envvar value
        monkeypatch.setattr(
            "bitfount.config.BITFOUNT_ENGINE", bitfount.config._BASIC_ENGINE
        )

        getGPUs = mocker.patch("bitfount.config.GPUtil.getGPUs")
        getGPUs.side_effect = Exception(
            "Pretend bad thing happened inside imported library"
        )

        gpu_metadata = bitfount.config.get_gpu_metadata()

        assert gpu_metadata == (None, 0)

    @pytest.mark.parametrize(
        "device_name,device_count,expected_result",
        [(None, 0, (None, 0)), ("BitfountGPU", 3, ("BitfountGPU", 3))],
    )
    def test_get_gpu_metadata_with_pytorch_engine(
        self,
        device_count: int,
        device_name: Optional[str],
        expected_result: Tuple[Optional[str], int],
        mocker: MockerFixture,
        monkeypatch: MonkeyPatch,
    ) -> None:
        """Test that CUDA is used to retrieve GPU info."""
        # Set envvar value
        monkeypatch.setattr(
            "bitfount.config.BITFOUNT_ENGINE", bitfount.config._PYTORCH_ENGINE
        )

        cuda = mocker.patch("bitfount.config.torch.cuda")
        cuda.get_device_name.return_value = device_name
        cuda.device_count.return_value = device_count

        gpu_metadata = bitfount.config.get_gpu_metadata()

        assert gpu_metadata == expected_result

    def test_get_gpu_metadata_with_pytorch_engine_throws_exception(
        self, mocker: MockerFixture, monkeypatch: MonkeyPatch
    ) -> None:
        """Test that GPUUtil is used to retrieve GPU info."""
        # Set envvar value
        monkeypatch.setattr(
            "bitfount.config.BITFOUNT_ENGINE", bitfount.config._PYTORCH_ENGINE
        )

        cuda = mocker.patch("bitfount.config.torch.cuda")
        cuda.get_device_name.side_effect = Exception(
            "Pretend bad thing happened inside imported library"
        )
        cuda.device_count.return_value = 0

        gpu_metadata = bitfount.config.get_gpu_metadata()

        assert gpu_metadata == (None, 0)


class TestBitfountEnvironment:
    """Tests that the BITFOUNT_ENVIRONMENT environment variable is read correctly."""

    @unit_test
    def test_get_environment_dev(self, monkeypatch: MonkeyPatch) -> None:
        """Tests dev environment is read correctly."""
        monkeypatch.setenv("BITFOUNT_ENVIRONMENT", "dev")
        env = _get_environment()
        assert env == _DEVELOPMENT_ENVIRONMENT

    @unit_test
    def test_get_environment_staging(self, monkeypatch: MonkeyPatch) -> None:
        """Tests staging environment is read correctly."""
        monkeypatch.setenv("BITFOUNT_ENVIRONMENT", "staging")
        env = _get_environment()
        assert env == _STAGING_ENVIRONMENT

    @unit_test
    def test_get_environment_production(self, monkeypatch: MonkeyPatch) -> None:
        """Tests production environment is read correctly."""
        monkeypatch.setenv("BITFOUNT_ENVIRONMENT", "production")
        env = _get_environment()
        assert env == _PRODUCTION_ENVIRONMENT

    @unit_test
    def test_get_environment_variable_not_set(self, monkeypatch: MonkeyPatch) -> None:
        """Tests that environment defaults to production variable has not been set."""
        monkeypatch.delenv("BITFOUNT_ENVIRONMENT", raising=False)
        env = _get_environment()
        assert env == _PRODUCTION_ENVIRONMENT

    @unit_test
    def test_get_environment_unknown_variable_raises_environment_error(
        self, monkeypatch: MonkeyPatch
    ) -> None:
        """Tests that error is raised if environment is not recognised."""
        monkeypatch.setenv("BITFOUNT_ENVIRONMENT", "not-a-real-env")
        with pytest.raises(ValueError):
            _get_environment()



"""Federated Averaging protocol."""
from __future__ import annotations

import asyncio
import os
import time
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Dict,
    List,
    Mapping,
    Optional,
    Protocol,
    Tuple,
    Union,
    cast,
    runtime_checkable,
)

from marshmallow import fields

from bitfount.data.datasources.base_source import BaseSource
from bitfount.federated.aggregators.base import (
    _AggregatorWorkerFactory,
    _BaseAggregator,
    _BaseAggregatorFactory,
    _BaseModellerAggregator,
    _BaseWorkerAggregator,
)
from bitfount.federated.aggregators.base import registry as aggregators_registry
from bitfount.federated.aggregators.secure import _InterPodAggregatorWorkerFactory
from bitfount.federated.algorithms.model_algorithms.base import (
    registry as algorithms_registry,
)
from bitfount.federated.early_stopping import FederatedEarlyStopping
from bitfount.federated.helper import _create_aggregator
from bitfount.federated.logging import _get_federated_logger
from bitfount.federated.model_reference import BitfountModelReference
from bitfount.federated.pod_vitals import _PodVitals
from bitfount.federated.privacy.differential import DPPodConfig
from bitfount.federated.protocols.base import (
    _BaseCompatibleAlgoFactory,
    _BaseModellerProtocol,
    _BaseProtocolFactory,
    _BaseWorkerProtocol,
)
from bitfount.federated.transport.modeller_transport import (
    _get_parameter_updates_from_workers,
    _get_training_metrics_from_workers,
    _ModellerMailbox,
    _send_model_parameters,
)
from bitfount.federated.transport.worker_transport import (
    _get_model_parameters,
    _InterPodWorkerMailbox,
    _send_parameter_update,
    _send_training_metrics,
    _WorkerMailbox,
)
from bitfount.federated.types import SerializedProtocol
from bitfount.hub.api import BitfountHub
from bitfount.schemas.utils import bf_dump
from bitfount.types import (
    T_DTYPE,
    T_FIELDS_DICT,
    T_NESTED_FIELDS,
    DistributedModelProtocol,
    _SerializedWeights,
    _Weights,
)
from bitfount.utils import delegates

if TYPE_CHECKING:
    from bitfount.types import _DistributedModelTypeOrReference

logger = _get_federated_logger(__name__)


@runtime_checkable
class _FederatedAveragingCompatibleAlgo(Protocol[T_DTYPE]):
    @property
    def epochs(self) -> Optional[int]:
        """Number of epochs."""
        ...

    @property
    def steps(self) -> Optional[int]:
        """Number of steps."""
        ...

    @property
    def tensor_precision(self) -> T_DTYPE:
        """Tensor precision."""
        ...


@runtime_checkable
class _FederatedAveragingCompatibleWorker(_FederatedAveragingCompatibleAlgo, Protocol):
    """Defines worker-side algorithm compatibility."""

    def initialise(
        self,
        datasource: BaseSource,
        pod_dp: Optional[DPPodConfig] = None,
        **kwargs: Any,
    ) -> None:
        """Initialises the worker-side algorithm."""
        ...

    def run(
        self,
        data: BaseSource,
        model_params: _SerializedWeights,
        iterations: int,
    ) -> Tuple[_Weights, Optional[Dict[str, str]]]:
        """Runs the worker-side algorithm."""
        ...

    def save_final_parameters(self, model_params: _SerializedWeights) -> None:
        """Saves the weights from the worker-side algorithm."""
        ...


@runtime_checkable
class _FederatedAveragingCompatibleModeller(
    _FederatedAveragingCompatibleAlgo, Protocol
):
    """Defines modeller-side algorithm compatibility."""

    def initialise(
        self,
        **kwargs: Any,
    ) -> None:
        """Initialises the modeller-side algorithm."""
        ...

    def run(
        self,
        update: Optional[_Weights] = None,
        validation_metrics: Optional[Mapping[str, float]] = None,
    ) -> _SerializedWeights:
        """Runs the modeller-side algorithm."""
        ...


@runtime_checkable
class _FederatedAveragingCompatibleAlgoFactory(_BaseCompatibleAlgoFactory, Protocol):
    """Defines algorithm factory compatibility."""

    model: _DistributedModelTypeOrReference
    pretrained_file: Optional[Union[str, os.PathLike]] = None

    def modeller(self, **kwargs: Any) -> _FederatedAveragingCompatibleModeller:
        """Returns a modeller-side algorithm."""
        ...

    def worker(
        self, hub: BitfountHub, **kwargs: Any
    ) -> _FederatedAveragingCompatibleWorker:
        """Returns a worker-side algorithm."""
        ...


class _BaseFederatedAveragingMixIn:
    """Shared behaviour for the `FederatedAveraging` classes."""

    # This is set in the base protocol
    algorithm: _FederatedAveragingCompatibleAlgo

    def __init__(
        self,
        *,
        aggregator: _BaseAggregator,
        steps_between_parameter_updates: Optional[int] = None,
        epochs_between_parameter_updates: Optional[int] = None,
        **_kwargs: Any,
    ):
        self.aggregator = aggregator
        self.steps_between_parameter_updates = steps_between_parameter_updates
        self.epochs_between_parameter_updates = epochs_between_parameter_updates

    def perform_iterations_checks(self) -> None:
        """Perform checks on iterations to ensure training configuration is correct.

        Raises:
            ValueError: if there is a mismatch between model iterations and
                algorithm iterations.
        """
        if bool(self.steps_between_parameter_updates) == bool(
            self.epochs_between_parameter_updates
        ):
            raise ValueError("You must specify one (and only one) of steps or epochs.")
        if bool(self.steps_between_parameter_updates) != bool(self.algorithm.steps):
            raise ValueError(
                "Parameter update method must match model training method"
                + " i.e. steps or epochs."
            )
        if (
            # If steps_between_parameter_updates then algorithm.steps is not None
            self.steps_between_parameter_updates
            and self.steps_between_parameter_updates > self.algorithm.steps  # type: ignore[operator] # Reason: see comment # noqa: B950
        ) or (
            # If epochs_between_parameter_updates then algorithm.epochs is not None
            self.epochs_between_parameter_updates
            and self.epochs_between_parameter_updates > self.algorithm.epochs  # type: ignore[operator] # Reason: see comment # noqa: B950
        ):
            raise ValueError(
                "Number of iterations between sharing updates must not be "
                "greater than total number of model iterations."
            )

    def get_num_federated_iterations(self) -> int:
        """Returns number of rounds of federated training to be done.

        This is rounded down to the nearest whole number.
        """
        num_iterations_between_updates = cast(
            int,
            self.epochs_between_parameter_updates
            or self.steps_between_parameter_updates,
        )
        num_iterations = cast(int, self.algorithm.epochs or self.algorithm.steps)

        # floor division rounds the result down to the nearest whole number
        return num_iterations // num_iterations_between_updates


class _ModellerSide(_BaseModellerProtocol, _BaseFederatedAveragingMixIn):
    """Modeller side of the FederatedAveraging protocol."""

    aggregator: _BaseModellerAggregator
    algorithm: _FederatedAveragingCompatibleModeller

    def __init__(
        self,
        *,
        algorithm: _FederatedAveragingCompatibleModeller,
        mailbox: _ModellerMailbox,
        aggregator: _BaseModellerAggregator,
        steps_between_parameter_updates: Optional[int],
        epochs_between_parameter_updates: Optional[int],
        early_stopping: Optional[FederatedEarlyStopping],
        auto_eval: bool = True,
        **kwargs: Any,
    ):
        super().__init__(
            algorithm=algorithm,
            mailbox=mailbox,
            aggregator=aggregator,
            steps_between_parameter_updates=steps_between_parameter_updates,
            epochs_between_parameter_updates=epochs_between_parameter_updates,
            **kwargs,
        )

        self.early_stopping = early_stopping
        self.auto_eval = auto_eval
        self.validation_results: List[Dict[str, float]] = []

    async def _send_parameters(self, new_parameters: _SerializedWeights) -> None:
        """Sends central model parameters to workers."""
        logger.debug("Sending global parameters to workers")
        await _send_model_parameters(new_parameters, self.mailbox)

    async def _receive_parameter_updates(
        self,
    ) -> Dict[str, _SerializedWeights]:
        """Receives parameter updates from every worker.

        Returns:
            A dictionary of the form {worker_name: weight_update}.
        """
        return await _get_parameter_updates_from_workers(self.mailbox)

    async def _get_training_metrics_updates(
        self,
    ) -> Dict[str, float]:
        return await _get_training_metrics_from_workers(self.mailbox)

    async def run(
        self,
        **kwargs: Any,
    ) -> List[Dict[str, float]]:
        """Receives updates and sends new parameters in a loop."""
        self.algorithm.initialise()
        self.perform_iterations_checks()
        initial_parameters = self.algorithm.run(update=None)
        await self._send_parameters(initial_parameters)
        num_federated_iterations = self.get_num_federated_iterations()
        for i in range(1, num_federated_iterations + 1):
            if self.algorithm.epochs:
                logger.info(f"Federated Epoch {i}")
            else:
                logger.info(f"Federated Step {i}")

            logger.info("Sending model parameters to Pods")
            if self.auto_eval:
                # We create this as a task so that it can process TRAINING_METRICS
                # messages in the background without blocking TRAINING_UPDATE
                # messages.
                validation_metrics_task = asyncio.create_task(
                    self._get_training_metrics_updates()
                )

            # from worker(s)
            weight_updates = await self._receive_parameter_updates()
            parameter_update = self.aggregator.run(
                algorithm_outputs=weight_updates,
                tensor_dtype=self.algorithm.tensor_precision,
            )
            if self.auto_eval:
                # This is guaranteed to be bound as it's creation is also in a
                # `if self.auto_eval:` block.
                # noinspection PyUnboundLocalVariable
                await validation_metrics_task
                validation_metrics: Dict[str, float] = validation_metrics_task.result()
                logger.info(
                    f"Validation Metrics at iteration {i}: {validation_metrics}"
                )
                # Each item in the list is the average results from every worker
                # for a given iteration. New results are appended to the list
                # such that the final item is always the latest.
                self.validation_results.append(validation_metrics)

                new_parameters = self.algorithm.run(
                    update=parameter_update, validation_metrics=validation_metrics
                )
                # Send the latest averaged validation metrics only at each iteration
            else:
                new_parameters = self.algorithm.run(update=parameter_update)

            logger.federated_info("Sending updated parameters")
            await self._send_parameters(
                new_parameters
            )  # Workers end up with final model

            # TODO: [BIT-970] consider moving early stopping to be handled in a side
            #       channel as part of handler based approach
            training_complete = False
            if self.early_stopping is not None:
                training_complete = self.early_stopping.check(self.validation_results)
            await self.mailbox.send_training_iteration_complete_update(
                training_complete
            )
            if training_complete:
                logger.info("Early stopping criterion met. Stopping training.")
                break
        await self.mailbox.send_task_complete_message()
        modeller_results = self.validation_results
        return modeller_results


class _WorkerSide(_BaseWorkerProtocol, _BaseFederatedAveragingMixIn):
    """Worker side of the FederatedAveraging protocol."""

    aggregator: _BaseWorkerAggregator
    algorithm: _FederatedAveragingCompatibleWorker

    def __init__(
        self,
        *,
        algorithm: _FederatedAveragingCompatibleWorker,
        mailbox: _WorkerMailbox,
        aggregator: _BaseWorkerAggregator,
        steps_between_parameter_updates: Optional[int],
        epochs_between_parameter_updates: Optional[int],
        auto_eval: bool = True,
        **kwargs: Any,
    ):
        super().__init__(
            algorithm=algorithm,
            mailbox=mailbox,
            aggregator=aggregator,
            steps_between_parameter_updates=steps_between_parameter_updates,
            epochs_between_parameter_updates=epochs_between_parameter_updates,
            auto_eval=auto_eval,
            **kwargs,
        )
        self.auto_eval = auto_eval

    async def _receive_parameters(self) -> _SerializedWeights:
        """Receives new global model parameters."""
        logger.debug("Receiving global parameters")
        return await _get_model_parameters(self.mailbox)

    async def _send_training_metrics(
        self,
        validation_metrics: Optional[Mapping[str, str]] = None,
    ) -> None:
        """Sends training metrics update."""
        if validation_metrics:
            logger.debug("Sending validation metrics to modeller")
            await _send_training_metrics(validation_metrics, self.mailbox)

    async def _send_parameter_update(
        self, parameter_update: _SerializedWeights
    ) -> None:
        """Sends parameter update."""
        logger.debug("Sending parameter update to modeller")
        await _send_parameter_update(parameter_update, self.mailbox)

    async def run(
        self,
        datasource: BaseSource,
        pod_dp: Optional[DPPodConfig] = None,
        pod_vitals: Optional[_PodVitals] = None,
        pod_identifier: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        """Receives parameters and sends updates in a loop."""
        self.algorithm.initialise(datasource=datasource, pod_dp=pod_dp)
        self.perform_iterations_checks()
        logger.debug("Waiting on initial parameters")
        serialized_model_params = await self._receive_parameters()
        logger.debug("Received initial parameters")

        num_federated_iterations = self.get_num_federated_iterations()
        for i in range(1, num_federated_iterations + 1):
            if self.algorithm.epochs:
                logger.info(f"Federated Epoch {i}")
                iterations = self.epochs_between_parameter_updates
            else:
                logger.info(f"Federated Step {i}")
                iterations = self.steps_between_parameter_updates
            iterations = cast(int, iterations)
            logger.federated_info("Running algorithm")
            if pod_vitals:
                pod_vitals.last_task_execution_time = time.time()
            parameter_update, validation_metrics = self.algorithm.run(
                datasource, serialized_model_params, iterations
            )

            # Aggregator returns the difference between the new and old model parameters
            aggregated_parameter_update = await self.aggregator.run(parameter_update)

            if self.auto_eval:
                logger.info(
                    f"Validation Metrics at iteration {i}: {validation_metrics}"
                )
                await self._send_training_metrics(validation_metrics)
            await self._send_parameter_update(aggregated_parameter_update)
            serialized_model_params = await self._receive_parameters()

            # TODO: [BIT-970] consider moving early stopping to be handled in a side
            # channel as part of handler based approach
            training_complete = (
                await self.mailbox.get_training_iteration_complete_update()
            )
            if training_complete:
                logger.info(
                    "Modeller reporting early stopping criterion met. "
                    + "Stopping training.",
                )
                break
        await self.mailbox.get_task_complete_update()
        self.algorithm.save_final_parameters(serialized_model_params)


@delegates()
class FederatedAveraging(_BaseProtocolFactory):
    """Original Federated Averaging algorithm by McMahan et al. (2017).

    This protocol performs a predetermined number of epochs or steps of training on
    each remote Pod before sending the updated model parameters to the modeller. These
    parameters are then averaged and sent back to the Pods for as many federated
    iterations as the Modeller specifies.

    :::tip

    For more information, take a look at the seminal paper:
    https://arxiv.org/abs/1602.05629

    :::

    Args:
        algorithm: The algorithm to use for training. This must be compatible with the
            `FederatedAveraging` protocol.
        aggregator: The aggregator to use for updating the model parameters across all
            Pods participating in the task. This argument takes priority over the
            `secure_aggregation` argument.
        steps_between_parameter_updates: The number of steps between parameter updates,
            i.e. the number of rounds of local training before parameters are updated.
            If `epochs_between_parameter_updates` is provided,
            `steps_between_parameter_updates` cannot be provided. Defaults to None.
        epochs_between_parameter_updates: The number of epochs between parameter
            updates, i.e. the number of rounds of local training before parameters are
            updated. If `steps_between_parameter_updates` is provided,
            `epochs_between_parameter_updates` cannot be provided. Defaults to None.
        auto_eval: Whether to automatically evaluate the model on the validation
            dataset. Defaults to True.
        secure_aggregation: Whether to use secure aggregation. This argument is
            overridden by the `aggregator` argument. Defaults to False.

    Attributes:
        name: The name of the protocol.
        algorithm: The algorithm to use for training
        aggregator: The aggregator to use for updating the model parameters.
        steps_between_parameter_updates: The number of steps between parameter updates.
        epochs_between_parameter_updates: The number of epochs between parameter
            updates.
        auto_eval: Whether to automatically evaluate the model on the validation
            dataset.

    Raises:
        TypeError: If the `algorithm` is not compatible with the protocol.
    """

    algorithm: _FederatedAveragingCompatibleAlgoFactory
    fields_dict: ClassVar[T_FIELDS_DICT] = {
        "steps_between_parameter_updates": fields.Integer(allow_none=True),
        "epochs_between_parameter_updates": fields.Integer(allow_none=True),
        "auto_eval": fields.Boolean(),
    }
    nested_fields: ClassVar[T_NESTED_FIELDS] = {
        "algorithm": algorithms_registry,
        "aggregator": aggregators_registry,
    }

    def __init__(
        self,
        *,
        algorithm: _FederatedAveragingCompatibleAlgoFactory,
        aggregator: Optional[_BaseAggregatorFactory] = None,
        steps_between_parameter_updates: Optional[int] = None,
        epochs_between_parameter_updates: Optional[int] = None,
        auto_eval: bool = True,
        secure_aggregation: bool = False,
        **kwargs: Any,
    ) -> None:
        if kwargs.get("model"):
            logger.warning("Ignoring provided model. Algorithm already has a model.")
        super().__init__(algorithm=algorithm, **kwargs)
        if aggregator:
            self.aggregator = aggregator
            logger.warning(
                "Aggregator provided, ignoring 'secure_aggregation' argument."
            )
        else:
            self.aggregator = _create_aggregator(secure_aggregation=secure_aggregation)
        self.steps_between_parameter_updates = steps_between_parameter_updates
        self.epochs_between_parameter_updates = epochs_between_parameter_updates
        self.auto_eval = auto_eval
        if not steps_between_parameter_updates and not epochs_between_parameter_updates:
            logger.info(
                "Neither steps_between_parameter_updates or "
                "epochs_between_parameter_updates were set ..."
            )
            if isinstance(algorithm.model, DistributedModelProtocol):
                if bool(algorithm.model.epochs):
                    logger.info("Setting epochs_between_parameter_updates to 1.")
                    self.epochs_between_parameter_updates = 1
                else:
                    logger.info("Setting steps_between_parameter_updates to 1.")
                    self.steps_between_parameter_updates = 1
            elif isinstance(self.algorithm.model, BitfountModelReference):
                if self.algorithm.model.hyperparameters.get("epochs"):
                    logger.info("Setting epochs_between_parameter_updates to 1.")
                    self.epochs_between_parameter_updates = 1
                else:
                    logger.info("Setting steps_between_parameter_updates to 1.")
                    self.steps_between_parameter_updates = 1

    @classmethod
    def _validate_algorithm(cls, algorithm: _BaseCompatibleAlgoFactory) -> None:
        """Checks that `algorithm` is compatible with the protocol.

        Raises:
            TypeError: If the `algorithm` is not compatible with the protocol.
        """
        if not isinstance(algorithm, _FederatedAveragingCompatibleAlgoFactory):
            raise TypeError(
                f"The {cls.__name__} protocol does not support "
                + f"the {type(algorithm).__name__} algorithm.",
            )

    def modeller(
        self,
        mailbox: _ModellerMailbox,
        early_stopping: Optional[FederatedEarlyStopping] = None,
        **kwargs: Any,
    ) -> _ModellerSide:
        """Returns the modeller side of the FederatedAveraging protocol."""
        return _ModellerSide(
            algorithm=self.algorithm.modeller(
                pretrained_file=self.algorithm.pretrained_file
            ),
            aggregator=self.aggregator.modeller(),
            steps_between_parameter_updates=self.steps_between_parameter_updates,
            epochs_between_parameter_updates=self.epochs_between_parameter_updates,
            auto_eval=self.auto_eval,
            mailbox=mailbox,
            early_stopping=early_stopping,
            **kwargs,
        )

    def worker(
        self,
        mailbox: _WorkerMailbox,
        hub: BitfountHub,
        **kwargs: Any,
    ) -> _WorkerSide:
        """Returns the worker side of the FederatedAveraging protocol.

        Raises:
            TypeError: If the mailbox is not compatible with the aggregator.
        """
        if isinstance(self.aggregator, _AggregatorWorkerFactory):
            worker_agg = self.aggregator.worker()
        elif isinstance(self.aggregator, _InterPodAggregatorWorkerFactory):
            if not isinstance(mailbox, _InterPodWorkerMailbox):
                raise TypeError(
                    "Inter-pod aggregators require an inter-pod worker mailbox."
                )
            worker_agg = self.aggregator.worker(mailbox=mailbox)
        else:
            raise TypeError(
                f"Unrecognised aggregator factory ({type(self.aggregator)}); "
                f"unable to determine how to call .worker() factory method."
            )
        return _WorkerSide(
            algorithm=self.algorithm.worker(hub=hub),
            aggregator=worker_agg,
            steps_between_parameter_updates=self.steps_between_parameter_updates,
            epochs_between_parameter_updates=self.epochs_between_parameter_updates,
            auto_eval=self.auto_eval,
            mailbox=mailbox,
            **kwargs,
        )

    def dump(self) -> SerializedProtocol:
        """Returns the JSON-serializable representation of the protocol."""
        return cast(SerializedProtocol, bf_dump(self))



"""Tests for differential privacy module."""
import copy
import logging

from _pytest.logging import LogCaptureFixture
import desert
import marshmallow
import numpy as np
import pytest
from pytest import fixture

from bitfount.federated.exceptions import DPParameterError
from bitfount.federated.privacy.differential import (
    DPModellerConfig,
    DPPodConfig,
    _DifferentiallyPrivate,
)
from bitfount.types import _StrAnyDict
from tests.utils.helper import unit_test


@fixture
def config_dict() -> _StrAnyDict:
    """Configuration dictionary for Differential Privacy settings."""
    return {
        "max_grad_norm": 1.0,
        # TODO: [BIT-1614] Re-enable user-provided noise_multiplier after the
        #       post-update privacy violation issue is resolved.
        # "noise_multiplier": 0.0,
        "alphas": [0.0, 1.1],
        "delta": 0.01,
        "loss_reduction": "sum",
        "epsilon": 10.0,
        "auto_fix": False,
    }


@unit_test
class TestDPModellerConfig:
    """Tests for differential privacy config dataclass."""

    def test_loss_reduction_validation(self) -> None:
        """Tests config rejects invalid loss_reduction values."""
        with pytest.raises(ValueError, match="loss_reduction must be one of"):
            # noinspection PyTypeChecker
            DPModellerConfig(epsilon=1.0, loss_reduction="test")  # type: ignore[arg-type] # Reason: purpose of test # noqa: B950

    def test_schema_loads(self, config_dict: _StrAnyDict) -> None:
        """Tests config loading from data."""
        schema = desert.schema(DPModellerConfig)
        test_config = schema.load(config_dict)

        # Assertions for correct conversion
        # TODO: [BIT-1614] Re-enable user-provided noise_multiplier after the
        #       post-update privacy violation issue is resolved.
        # assert test_config.noise_multiplier == config_dict["noise_multiplier"]
        assert test_config.max_grad_norm == config_dict["max_grad_norm"]
        assert test_config.alphas == config_dict["alphas"]
        assert test_config.delta == config_dict["delta"]
        assert test_config.loss_reduction == config_dict["loss_reduction"]
        assert test_config.epsilon == config_dict["epsilon"]
        assert test_config.auto_fix == config_dict["auto_fix"]

    def test_schema_validation_failures(self) -> None:
        """Tests that schema validation catches incorrect options."""
        schema = desert.schema(DPModellerConfig)
        with pytest.raises(marshmallow.exceptions.ValidationError):
            schema.load({"loss_reduction": "test"})

    def test_error_alpha_1(self) -> None:
        """Tests error is raised when one alpha is 1."""
        with pytest.raises(DPParameterError):
            DPModellerConfig(
                epsilon=1,
                delta=1e-6,
                alphas=list(np.linspace(1.0, 440.0, 440)),
            )


@unit_test
class TestDifferentiallyPrivate:
    """Tests related to the DifferentiallyPrivate class."""

    @fixture
    def dp_config(self) -> DPModellerConfig:
        """Creates a DPConfig instance."""
        return DPModellerConfig(epsilon=10.0, delta=1.0)

    @fixture
    def pod_dp_config(self) -> DPPodConfig:
        """Creates a DPPodConfig instance."""
        return DPPodConfig(epsilon=1.0, delta=0.5)

    def test__convert_to_dpconfig(self, config_dict: _StrAnyDict) -> None:
        """Tests the conversion from dict to DPConfig."""
        test_config = _DifferentiallyPrivate._convert_to_dpconfig(config_dict)

        # Assertions for correct conversion
        assert test_config is not None
        # TODO: [BIT-1614] Re-enable user-provided noise_multiplier after the
        #       post-update privacy violation issue is resolved.
        # assert test_config.noise_multiplier == config_dict["noise_multiplier"]
        assert test_config.max_grad_norm == config_dict["max_grad_norm"]
        assert test_config.alphas == config_dict["alphas"]
        assert test_config.delta == config_dict["delta"]
        assert test_config.loss_reduction == config_dict["loss_reduction"]
        assert test_config.epsilon == config_dict["epsilon"]
        assert test_config.auto_fix == config_dict["auto_fix"]

    def test__convert_to_dpconfig_with_none(self) -> None:
        """Tests that None is returned if None supplied."""
        assert None is _DifferentiallyPrivate._convert_to_dpconfig(None)

    def test_config_settings_logged(
        self, caplog: LogCaptureFixture, dp_config: DPModellerConfig
    ) -> None:
        """Tests that config settings are logged out."""
        caplog.set_level(logging.INFO)
        _DifferentiallyPrivate(dp_config)
        assert str(dp_config) in caplog.text

    def test_no_config_supplied_is_logged(self, caplog: LogCaptureFixture) -> None:
        """Tests that lack of config settings is logged out."""
        caplog.set_level(logging.INFO)
        _DifferentiallyPrivate(None)
        assert "No differential privacy settings provided." in caplog.text

    def test_apply_pod_dp_no_modeller_config(self) -> None:
        """Tests apply_pod_dp causes no change if no modeller config."""
        dp = _DifferentiallyPrivate()
        dp.apply_pod_dp(None)
        assert dp._dp_config is None

    def test_apply_pod_dp_no_pod_config(
        self, caplog: LogCaptureFixture, dp_config: DPModellerConfig
    ) -> None:
        """Tests apply_pod_dp causes no changes if no pod dp."""
        caplog.set_level(logging.INFO)
        orig_dp_config = copy.deepcopy(dp_config)
        dp = _DifferentiallyPrivate(dp_config)
        dp.apply_pod_dp(None)

        assert vars(dp._dp_config) == vars(orig_dp_config)  # hasn't changed
        assert "No pod DP preferences, using modeller preferences." in caplog.text

    def test_apply_pod_dp_caps_values(
        self,
        caplog: LogCaptureFixture,
        dp_config: DPModellerConfig,
        pod_dp_config: DPPodConfig,
    ) -> None:
        """Tests that limits are applied and logged from pod DP config."""
        orig_dp_config = copy.deepcopy(dp_config)
        dp = _DifferentiallyPrivate(dp_config)
        dp.apply_pod_dp(pod_dp_config)

        # Check config applied
        assert dp._dp_config is not None

        # Check caps are applied
        assert dp._dp_config.epsilon == pod_dp_config.epsilon
        assert dp._dp_config.delta == pod_dp_config.delta

        # Check these are different from the supplied ones
        assert dp._dp_config.epsilon != orig_dp_config.epsilon
        assert dp._dp_config.delta != orig_dp_config.delta

        # Check these changes are logged
        assert (
            f"Requested DP max epsilon ({orig_dp_config.epsilon}) exceeds "
            f"maximum value allowed by pod. Using pod max of "
            f"{pod_dp_config.epsilon}." in caplog.text
        )
        assert (
            f"Requested DP target delta ({orig_dp_config.delta}) exceeds "
            f"maximum value allowed by pod. Using pod max of "
            f"{pod_dp_config.delta}." in caplog.text
        )



"""Transport layer classes for communicating tasks between pods and modellers."""

#: Maximum size of a message: 700MB.
MAXIMUM_GRPC_MESSAGE_SIZE_BYTES = 700 * 1024 * 1024
_MESSAGE_SERVICE_GRPC_OPTIONS = [
    ("grpc.max_message_length", MAXIMUM_GRPC_MESSAGE_SIZE_BYTES),
    ("grpc.max_receive_message_length", MAXIMUM_GRPC_MESSAGE_SIZE_BYTES),
]



"""Tests logging.py."""

import logging
from typing import Dict
from unittest.mock import Mock

from _pytest.logging import LogCaptureFixture
import pytest
from pytest_mock import MockerFixture

from bitfount.federated.logging import (
    _federate_logger,
    _FederatedLogFilter,
    _FederatedLogger,
    _get_federated_logger,
    _MailboxHandler,
)
from bitfount.federated.transport.base_transport import _BaseMailbox
from tests.utils.helper import unit_test

LOGGING_METHODS = ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]


@pytest.fixture
def log_message() -> str:
    """Returns dummy log message."""
    return "testing123"


@unit_test
@pytest.mark.parametrize("level", LOGGING_METHODS + ["notareallogginglevel"])
def test_get_federated_logger(caplog: LogCaptureFixture, level: str) -> None:
    """Tests `get_federated_logger` returns a logger with the right methods."""
    logger = _get_federated_logger("bitfount.federated.test")
    msg = f"{level}_msg"

    assert logger.name == "bitfount.federated.test"

    if level in LOGGING_METHODS:
        assert hasattr(logger, f"federated_{level.lower()}")
        method_to_call = getattr(logger, f"federated_{level.lower()}")

        # Confirming that the method can be called - an error will be raised if it can't
        method_to_call(msg)
    else:
        assert not hasattr(logger, f"federated_{level.lower()}")

    for record in caplog.records:
        assert record.levelname == level
        assert record.msg == msg


@unit_test
def test_federate_logger() -> None:
    """Tests that `federate_logger` has the right handlers and filters."""
    mock_mailbox = Mock()
    _federate_logger(mock_mailbox)

    federated_logger = logging.getLogger("bitfount.federated")

    # Make assertions on logger
    assert federated_logger.hasHandlers()
    assert len(federated_logger.handlers) == 1
    handler = federated_logger.handlers[0]
    assert isinstance(handler, _MailboxHandler)

    # Make sure that the filter is on the handler rather than the logger
    assert len(federated_logger.filters) == 0

    # Make assertions on handler
    assert handler.level == logging.DEBUG
    assert len(handler.filters) == 1
    assert isinstance(handler.filters[0], _FederatedLogFilter)


@unit_test
def test_federate_logger_only_attaches_handler_once() -> None:
    """Tests that federate_logger only attaches handler once."""
    mock_mailbox = Mock(mock_name="first-mailbox")
    _federate_logger(mock_mailbox)
    federated_logger = logging.getLogger("bitfount.federated")
    assert len(federated_logger.handlers) == 1
    handler = federated_logger.handlers[0]
    assert isinstance(handler, _MailboxHandler)
    assert handler.mailbox == mock_mailbox

    # We make a second call to `federate_logger` which removes the first handler
    mock_mailbox2 = Mock(mock_name="second-mailbox")
    _federate_logger(mock_mailbox2)
    assert len(federated_logger.handlers) == 1
    handler = federated_logger.handlers[0]
    assert isinstance(handler, _MailboxHandler)
    assert handler.mailbox == mock_mailbox2


@unit_test
@pytest.mark.parametrize(
    "name",
    [
        "bitfount.federated",
        "bitfount.federated.aggregators",
        "bitfount.federated.protocols.base",
        "bitfount.federated.transport.base_transport",
        "tests.bitfount.federated",
        "bitfount.hub",
    ],
)
def test_federated_logger_propagates(log_message: str, name: str) -> None:
    """Tests that module loggers within the federated package also become federated.

    Those not in the `federated` package raise a ValueError.
    """
    mock_mailbox = Mock()
    _federate_logger(mock_mailbox)
    if not name.startswith("bitfount.federated"):
        with pytest.raises(ValueError):
            _get_federated_logger(name)
    else:
        logger = _get_federated_logger(name)
        assert isinstance(logger, _FederatedLogger)
        logger.warning(log_message)
        logger.federated_warning(name)
        mock_mailbox.log.assert_called_once()
        assert mock_mailbox.log.call_args[0][0]["msg"] == name


@unit_test
@pytest.mark.parametrize(
    "log_method, extra_arguments, federated",
    [
        ("warning", {}, False),
        ("warning", {"federated": False}, False),
        ("warning", {"federated": True}, True),
        ("federated_warning", {}, True),
    ],
)
def test_federated_logger_filter(
    caplog: LogCaptureFixture,
    extra_arguments: Dict[str, bool],
    federated: bool,
    log_message: str,
    log_method: str,
) -> None:
    """Tests the the FederatedLogFilter correctly filters messages prior to the Handler.

    Tests that only log messages with the `federated` attribute set to True are passed
    on to the MailboxHandler.
    """
    mock_mailbox = Mock()
    _federate_logger(mock_mailbox)
    logger = _get_federated_logger("bitfount.federated")
    method_to_call = getattr(logger, log_method)
    if extra_arguments:
        method_to_call(log_message, extra=extra_arguments)
    else:
        method_to_call(log_message)

    assert len(caplog.records) == 1
    record = caplog.records[0]

    assert record.levelname == "WARNING"
    assert record.msg == log_message

    if federated:
        mock_mailbox.log.assert_called_once()
    else:
        mock_mailbox.log.assert_not_called()


@pytest.mark.parametrize("levelno", [10, 20, 30, 40, 50])
@unit_test
def test_emit_method_creates_asyncio_task(levelno: int, mocker: MockerFixture) -> None:
    """Tests that the emit method creates an asyncio task.

    It also tests that messages with a levelno greater than or equal to `ERROR` are sent
    with priority.
    """
    mock_mailbox = Mock(spec=_BaseMailbox)
    handler = _MailboxHandler(mock_mailbox)
    record = Mock(spec=logging.LogRecord)
    record.levelno = levelno
    mock_asyncio = mocker.patch("bitfount.federated.logging.asyncio")
    handler.emit(record)

    # Assert that asyncio task is created
    mock_asyncio.create_task.assert_called_once()

    # Assert that messages of type ERROR or higher are sent with priority on the event
    # loop
    if levelno >= logging.ERROR:
        mock_asyncio.get_running_loop.assert_called_once()
    else:
        mock_asyncio.get_running_loop.assert_not_called()



"""Transport layer exceptions."""
from bitfount.exceptions import BitfountError


class BitfountMessageServiceError(BitfountError):
    """Bitfount exception related to interaction with the message service."""

    pass


class TaskRequestError(BitfountError):
    """Bitfount exception related to the task request."""

    pass
